"""
pygame-menu
https://github.com/ppizarror/pygame-menu

BUTTON
Button class, manage elements and adds entries to Menu.
"""

__all__ = [
    'Button',
    'ButtonManager'
]

import pygame
import pygame_menu
import pygame_menu.events as _events
import re
import webbrowser

from abc import ABC
from pygame_menu.locals import FINGERUP, CURSOR_HAND
from pygame_menu.utils import assert_color, get_finger_pos, warn
from pygame_menu.widgets.core.widget import AbstractWidgetManager, Widget

from pygame_menu._types import Any, CallbackType, Callable, Union, List, Tuple, \
    Optional, ColorType, ColorInputType, EventVectorType


# noinspection PyMissingOrEmptyDocstring
class Button(Widget):
    """
    Button widget.

    The arguments and unknown keyword arguments are passed to the ``onreturn``
    function:

    .. code-block:: python

        onreturn(*args, **kwargs)

    .. note::

        Button accepts all transformations.

    :param title: Button title
    :param button_id: Button ID
    :param onreturn: Callback when pressing the button
    :param args: Optional arguments for callbacks
    :param kwargs: Optional keyword arguments
    """
    _last_underline: List[Union[str, Optional[Tuple[ColorType, int, int]]]]  # deco id, (color, offset, width)
    to_menu: bool

    def __init__(
        self,
        title: Any,
        button_id: str = '',
        onreturn: CallbackType = None,
        *args,
        **kwargs
    ) -> None:
        super(Button, self).__init__(
            args=args,
            kwargs=kwargs,
            onreturn=onreturn,
            title=title,
            widget_id=button_id
        )
        self._accept_events = True
        self._last_underline = ['', None]
        self.to_menu = False  # True if the button opens a new Menu

    def _apply_font(self) -> None:
        pass

    def set_selection_callback(
        self,
        callback: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]]
    ) -> None:
        """
        Update the button selection callback, once button is selected, the callback
        function is executed as follows:

        .. code-block:: python

            callback(selected, widget, menu)

        :param callback: Callback when selecting the widget, executed in :py:meth:`pygame_menu.widgets.core.widget.Widget.set_selected`
        """
        if callback is not None:
            assert callable(callback), \
                'callback must be callable (function-type) or None'
        self._onselect = callback

    def update_callback(self, callback: Callable, *args) -> None:
        """
        Update function triggered by the button; ``callback`` cannot point to a Menu, that
        behaviour is only valid using :py:meth:`pygame_menu.menu.Menu.add.button` method.

        .. note::

            If button points to a submenu, and the callback is changed to a
            function, the submenu will be removed from the parent Menu. Thus
            preserving the structure.

        :param callback: Function
        :param args: Arguments used by the function once triggered
        """
        assert callable(callback), \
            'only callable (function-type) are allowed'

        # If return is a Menu object, remove it from submenus list
        if self._menu is not None and self._onreturn is not None and self.to_menu:
            assert len(self._args) == 1
            submenu = self._args[0]  # Menu
            assert self._menu.in_submenu(submenu), \
                'pointed menu is not in submenu list of parent container'
            # noinspection PyProtectedMember
            assert self._menu._remove_submenu(submenu, self), \
                'submenu could not be removed'
            self.to_menu = False

        self._args = args or []
        self._onreturn = callback

    def add_underline(
        self,
        color: ColorInputType,
        offset: int,
        width: int,
        force_render: bool = False
    ) -> 'Button':
        """
        Adds an underline to text. This is added if widget is rendered.

        :param color: Underline color
        :param offset: Underline offset
        :param width: Underline width
        :param force_render: If ``True`` force widget render after addition
        :return: Self reference
        """
        color = assert_color(color)
        assert isinstance(offset, int)
        assert isinstance(width, int) and width > 0
        self._last_underline[1] = (color, offset, width)
        if force_render:
            self._force_render()
        return self

    def remove_underline(self) -> 'Button':
        """
        Remove underline of the button.

        :return: Self reference
        """
        if self._last_underline[0] != '':
            self._decorator.remove(self._last_underline[0])
            self._last_underline[0] = ''
        return self

    def _draw(self, surface: 'pygame.Surface') -> None:
        surface.blit(self._surface, self._rect.topleft)

    def _render(self) -> Optional[bool]:
        if not self._render_hash_changed(self._selected, self._title, self._visible, self.readonly, self._last_underline[1]):
            return True

        # Render surface
        self._surface = self._render_string(self._title, self.get_font_color_status())
        self._apply_transforms()
        self._rect.width, self._rect.height = self._surface.get_size()

        # Add underline if enabled
        self.remove_underline()
        if self._last_underline[1] is not None:
            w = self._surface.get_width()
            h = self._surface.get_height()
            color, offset, width = self._last_underline[1]
            if w > 0 and h > 0:
                self._last_underline[0] = self._decorator.add_line(
                    pos1=(-w / 2, h / 2 + offset),
                    pos2=(w / 2, h / 2 + offset),
                    color=color,
                    width=width
                )

        self.force_menu_surface_update()

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)
        rect = self.get_rect(to_real_position=True)

        if self.readonly or not self.is_visible():
            self._readonly_check_mouseover(events, rect)
            return False

        for event in events:

            # Check mouse over
            self._check_mouseover(event, rect)

            # User applies with key
            if (
                event.type == pygame.KEYDOWN and self._keyboard_enabled and self._ctrl.apply(event, self) or
                event.type == pygame.JOYBUTTONDOWN and self._joystick_enabled and self._ctrl.joy_select(event, self)
            ):
                if self.to_menu:
                    self._sound.play_open_menu()
                else:
                    self._sound.play_key_add()
                self.apply()
                return True

            # User clicks the button; don't consider the mouse wheel (button 4 & 5)
            elif (
                event.type == pygame.MOUSEBUTTONUP and self._mouse_enabled and event.button in (1, 2, 3) or
                event.type == FINGERUP and self._touchscreen_enabled and self._menu is not None
            ):
                if event.type == pygame.MOUSEBUTTONUP:
                    self._sound.play_click_mouse()
                else:
                    self._sound.play_click_touch()

                event_pos = get_finger_pos(self._menu, event)
                if rect.collidepoint(*event_pos):
                    self.apply()
                    return True

        return False


class ButtonManager(AbstractWidgetManager, ABC):
    """
    Button manager.
    """

    # noinspection PyProtectedMember
    def banner(
        self,
        image: 'pygame_menu.BaseImage',
        action: Optional[Union['pygame_menu.Menu', '_events.MenuAction', Callable, int]] = None,
        *args,
        **kwargs
    ) -> 'pygame_menu.widgets.Button':
        """
        Adds a clickeable image to the Menu with same behaviour as a Button.

        The arguments and unknown keyword arguments are passed to the action, if
        it's a callable object:

        .. code-block:: python

            action(*args)

        If ``accept_kwargs=True`` then the ``**kwargs`` are also unpacked on action
        call:

        .. code-block:: python

            action(*args, **kwargs)

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``accept_kwargs``                 (bool) – Button action accepts ``**kwargs`` if it's a callable object (function-type), ``False`` by default
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``back_count``                    (int) – Number of menus to go back if action is :py:data:`pygame_menu.events.BACK` event, default is ``1``
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``button_id``                     (str) – Widget ID
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``onselect``                      (callable, None) – Callback executed when selecting the widget
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled

        .. note::

            All theme-related optional kwargs use the default Menu theme if not defined.

        .. note::

            Using ``action=None`` is the same as using ``action=pygame_menu.events.NONE``.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Be careful with kwargs collision. Consider that all optional documented
            kwargs keys are removed from the object.

        :param image: Image of the clickeable button
        :param action: Action of the button, can be a Menu, an event, or a function
        :param args: Additional arguments used by a function
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.Button`
        """
        kwargs['background_color'] = image
        button = self.button(' ', action, *args, **kwargs)
        return button.resize(*image.get_size())

    # noinspection PyProtectedMember
    def button(
        self,
        title: Any,
        action: Optional[Union['pygame_menu.Menu', '_events.MenuAction', Callable, int]] = None,
        *args,
        **kwargs
    ) -> 'pygame_menu.widgets.Button':
        """
        Adds a button to the Menu.

        The arguments and unknown keyword arguments are passed to the action, if
        it's a callable object:

        .. code-block:: python

            action(*args)

        If ``accept_kwargs=True`` then the ``**kwargs`` are also unpacked on action
        call:

        .. code-block:: python

            action(*args, **kwargs)

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``accept_kwargs``                 (bool) – Button action accepts ``**kwargs`` if it's a callable object (function-type), ``False`` by default
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``back_count``                    (int) – Number of menus to go back if action is :py:data:`pygame_menu.events.BACK` event, default is ``1``
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``button_id``                     (str) – Widget ID
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``onselect``                      (callable, None) – Callback executed when selecting the widget
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``readonly_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode
            - ``readonly_selected_color``       (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode and is selected
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``tab_size``                      (int) – Width of a tab character
            - ``underline_color``               (tuple, list, str, int, :py:class:`pygame.Color`, None) – Color of the underline. If ``None`` use the same color of the text
            - ``underline_offset``              (int) – Vertical offset in px. ``2`` by default
            - ``underline_width``               (int) – Underline width in px. ``2`` by default
            - ``underline``                     (bool) – Enables text underline, using a properly placed decoration. ``False`` by default

        .. note::

            All theme-related optional kwargs use the default Menu theme if not defined.

        .. note::

            Using ``action=None`` is the same as using ``action=pygame_menu.events.NONE``.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Be careful with kwargs collision. Consider that all optional documented
            kwargs keys are removed from the object.

        :param title: Title of the button
        :param action: Action of the button, can be a Menu, an event, or a function
        :param args: Additional arguments used by a function
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.Button`
        """
        total_back = kwargs.pop('back_count', 1)
        assert isinstance(total_back, int) and 1 <= total_back

        # Get ID
        button_id = kwargs.pop('button_id', '')
        assert isinstance(button_id, str), 'id must be a string'

        # Accept kwargs
        accept_kwargs = kwargs.pop('accept_kwargs', False)
        assert isinstance(accept_kwargs, bool)

        # Onselect callback
        onselect = kwargs.pop('onselect', None)

        # Filter widget attributes to avoid passing them to the callbacks
        attributes = self._filter_widget_attributes(kwargs)

        # Button underline
        underline = kwargs.pop('underline', False)
        underline_color = kwargs.pop('underline_color', attributes['font_color'])
        underline_offset = kwargs.pop('underline_offset', 1)
        underline_width = kwargs.pop('underline_width', 1)

        # Change action if certain events
        if action == _events.PYGAME_QUIT or action == _events.PYGAME_WINDOWCLOSE:
            action = _events.EXIT
        elif action is None:
            action = _events.NONE

        # If element is a Menu
        if isinstance(action, type(self._menu)):
            # Check for recursive
            if action == self._menu or action.in_submenu(self._menu, recursive=True):
                raise ValueError(
                    f'{action.get_class_id()} title "{action.get_title()}" is '
                    f'already on submenu structure, recursive menus lead to '
                    f'unexpected behaviours. For returning to previous menu'
                    f'use pygame_menu.events.BACK event defining an optional '
                    f'back_count number of menus to return from, default is 1'
                )

            widget = Button(title, button_id, self._menu._open, action)
            widget.to_menu = True

        # If element is a MenuAction
        elif action == _events.BACK:  # Back to Menu
            widget = Button(title, button_id, self._menu.reset, total_back)

        elif action == _events.CLOSE:  # Close Menu
            widget = Button(title, button_id, self._menu._close)

        elif action == _events.EXIT:  # Exit program
            widget = Button(title, button_id, self._menu._exit)

        elif action == _events.NONE:  # None action
            widget = Button(title, button_id)

        elif action == _events.RESET:  # Back to Top Menu
            widget = Button(title, button_id, self._menu.full_reset)

        # If element is a function or callable
        elif callable(action):
            if not accept_kwargs:
                widget = Button(title, button_id, action, *args)
            else:
                widget = Button(title, button_id, action, *args, **kwargs)

        else:
            raise ValueError('action must be a Menu, a MenuAction (event), a '
                             'function (callable), or None')

        # Configure and add the button
        if not accept_kwargs:
            try:
                self._check_kwargs(kwargs)
            except ValueError:
                if self._menu._verbose:
                    warn('button cannot accept kwargs. If you want to use kwargs '
                         'options set accept_kwargs=True')
                raise

        self._configure_widget(widget=widget, **attributes)
        if underline:
            widget.add_underline(underline_color, underline_offset, underline_width)
        widget.set_selection_callback(onselect)
        self._append_widget(widget)

        # Add to submenu
        if widget.to_menu:
            self._add_submenu(action, widget)

        return widget

    def url(
        self,
        href: str,
        title: str = '',
        **kwargs
    ) -> 'pygame_menu.widgets.Button':
        """
        Adds a Button url to the Menu. Clicking the widget will open the link.
        If ``title`` is defined, the link will not be written. For example:
        ``href='google.com', title=''`` will write the link, but
        ``href='google.com', title='Google'`` will write 'Google' and opens
        'google.com' if clicked.

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over. By default, is ``HAND``
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color. If not defined, uses ``theme.widget_url_color``
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``tab_size``                      (int) – Width of a tab character
            - ``underline_color``               (tuple, list, str, int, :py:class:`pygame.Color`, None) – Color of the underline. If ``None`` use the same color of the text
            - ``underline_offset``              (int) – Vertical offset in px. ``2`` by default
            - ``underline_width``               (int) – Underline width in px. ``2`` by default
            - ``underline``                     (bool) – Enables text underline, using a properly placed decoration. ``True`` by default

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param href: Link to open
        :param title: Alternative title of the link
        :param kwargs: Optional keyword arguments
        :return: Widget object, or List of widgets if the text overflows
        :rtype: :py:class:`pygame_menu.widgets.Button`
        """
        # Validate link
        assert isinstance(href, str) and len(href) > 0

        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z\d](?:[A-Z\d-]{0,61}[A-Z\d])?\.)+(?:[A-Z]{2,6}\.?|[A-Z\d-]{2,}\.?)|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)
        assert re.match(regex, href) is not None, 'invalid link format'

        # Configure kwargs
        if 'cursor' not in kwargs.keys():
            kwargs['cursor'] = CURSOR_HAND
        if 'font_color' not in kwargs.keys():
            kwargs['font_color'] = self._theme.widget_url_color
        if 'selection_color' not in kwargs.keys():
            kwargs['selection_color'] = self._theme.widget_url_color
        if 'selection_effect' not in kwargs.keys():
            kwargs['selection_effect'] = pygame_menu.widgets.NoneSelection()
        if 'underline' not in kwargs.keys():
            kwargs['underline'] = True

        # Return new button
        return self.button(title if title != '' else href, lambda: webbrowser.open(href), **kwargs)



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

DROPSELECT
Drop select widget. This is similar to HTML selects, it can contain many items
(options) to select. The selection is unique.
"""

__all__ = [
    'DropSelect',
    'DropSelectManager'
]

import math
import pygame
import pygame_menu

from abc import ABC
from pygame_menu.font import FontType, get_font, assert_font
from pygame_menu.locals import ORIENTATION_VERTICAL, FINGERDOWN, FINGERUP, \
    POSITION_NORTHWEST, POSITION_SOUTHEAST
from pygame_menu.utils import check_key_pressed_valid, assert_color, assert_vector, \
    make_surface, parse_padding, get_finger_pos, uuid4, assert_cursor, assert_position
from pygame_menu.widgets.core.widget import AbstractWidgetManager, Widget, \
    WidgetTransformationNotImplemented
from pygame_menu.widgets.widget.button import Button
from pygame_menu.widgets.widget.frame import Frame
from pygame_menu.widgets.widget.selector import check_selector_items

from pygame_menu._types import Tuple, Union, List, Any, Optional, CallbackType, \
    ColorType, Dict, ColorInputType, Tuple2IntType, Tuple3IntType, PaddingType, \
    PaddingInstance, Tuple4IntType, NumberType, EventVectorType, Tuple2NumberType, \
    CursorInputType, CursorType, NumberInstance, Callable


# noinspection PyMissingOrEmptyDocstring,PyProtectedMember
class DropSelect(Widget):
    """
    Drop select is a selector within a Frame. This drops a vertical frame if
    requested. Drop select can contain selectable items (options), only 1 can be
    selected.

    The items of the DropSelect are:

    .. code-block:: python

        items = [('Item1', a, b, c...), ('Item2', d, e, f...)]

    The callbacks receive the current selected item, its index in the list, the
    associated arguments, and all unknown keyword arguments, where
    ``selected_item=widget.get_value()`` and ``selected_index=widget.get_index()``:

    .. code-block:: python

        onchange((selected_item, selected_index), a, b, c..., **kwargs)
        onreturn((selected_item, selected_index), a, b, c..., **kwargs)

    For example, if ``selected_index=0`` then ``selected_item=('Item1', a, b, c...)``.

    .. note::

        DropSelect only accepts translation transformation.

    :param title: Drop select title
    :param items: Items of the drop select
    :param dropselect_id: ID of the drop select
    :param default: Index of default item to display. If ``None`` no item is selected
    :param onchange: Callback when changing the drop select item
    :param onreturn: Callback when pressing return (apply) on the selected item
    :param onselect: Function when selecting the widget
    :param open_middle: If ``True`` the selection box is opened in the middle of the menu
    :param placeholder: Text shown if no option is selected yet
    :param placeholder_add_to_selection_box: If ``True`` adds the placeholder button to the selection box
    :param scrollbar_color: Scrollbar color
    :param scrollbar_cursor: Cursor of the scrollbars if mouse is placed over. By default, is ``None``
    :param scrollbar_shadow: Indicate if a shadow is drawn on each scrollbar
    :param scrollbar_shadow_color: Color of the shadow of each scrollbar
    :param scrollbar_shadow_offset: Offset of the scrollbar shadow in px
    :param scrollbar_shadow_position: Position of the scrollbar shadow. See :py:mod:`pygame_menu.locals`
    :param scrollbar_slider_color: Color of the sliders
    :param scrollbar_slider_hover_color: Color of the slider if hovered or clicked
    :param scrollbar_slider_pad: Space between slider and scrollbars borders in px
    :param scrollbar_thick: Scrollbar thickness in px
    :param scrollbars: Scrollbar position. See :py:mod:`pygame_menu.locals`
    :param selection_box_arrow_color: Selection box arrow color
    :param selection_box_arrow_margin: Selection box arrow margin (left, right, vertical) in px
    :param selection_box_bgcolor: Selection box background color
    :param selection_box_border_color: Selection box border color
    :param selection_box_border_width: Selection box border width
    :param selection_box_height: Selection box height, counted as how many options are packed before showing scroll
    :param selection_box_inflate: Selection box inflate on x-axis and y-axis (x, y) in px
    :param selection_box_margin: Selection box on x-axis and y-axis (x, y) margin from title in px
    :param selection_box_text_margin: Selection box text margin (left) in px
    :param selection_box_width: Selection box width in px. If ``0`` compute automatically to fit placeholder
    :param selection_infinite: If ``True`` selection can rotate through bottom/top
    :param selection_option_border_color: Option border color
    :param selection_option_border_width: Option border width
    :param selection_option_cursor: Option cursor. If ``None`` use the same cursor as the widget
    :param selection_option_font: Option font. If ``None`` use the same font as the widget
    :param selection_option_font_color: Option font color
    :param selection_option_font_size: Option font size. If ``None`` use the 100% of the widget font size
    :param selection_option_padding: Selection padding. See padding styling
    :param selection_option_selected_bgcolor: Selected option background color
    :param selection_option_selected_font_color: Selected option font color
    :param kwargs: Optional keyword arguments
    """
    _close_on_apply: bool
    _drop_frame: Optional['Frame']
    _index: int
    _items: Union[List[Tuple[Any, ...]], List[str]]
    _open_bottom: bool
    _open_middle: bool
    _opened: bool
    _option_buttons: List['Button']
    _option_font: Optional['pygame.font.Font']
    _placeholder: str
    _placeholder_add_to_selection_box: bool
    _scrollbar_color: ColorType
    _scrollbar_cursor: CursorType
    _scrollbar_shadow: bool
    _scrollbar_shadow_color: ColorType
    _scrollbar_shadow_offset: int
    _scrollbar_shadow_position: str
    _scrollbar_slider_color: ColorType
    _scrollbar_slider_hover_color: ColorType
    _scrollbar_slider_pad: NumberType
    _scrollbar_thick: int
    _scrollbars: str
    _selection_box_arrow_color: ColorType
    _selection_box_arrow_margin: Tuple3IntType
    _selection_box_bgcolor: ColorType
    _selection_box_border_color: ColorType
    _selection_box_border_width: int
    _selection_box_height: int
    _selection_box_inflate: Tuple2IntType
    _selection_box_margin: Tuple2IntType
    _selection_box_text_margin: int
    _selection_box_width: int
    _selection_infinite: bool
    _selection_option_border_color: ColorType
    _selection_option_border_width: int
    _selection_option_cursor: CursorType
    _selection_option_font_style: Dict[str, Any]
    _selection_option_left_space: bool
    _selection_option_left_space_height_factor: float
    _selection_option_left_space_margin: Tuple3IntType
    _selection_option_padding: Tuple4IntType
    _selection_option_selected_bgcolor: ColorType
    _title_size: Tuple2IntType

    def __init__(
        self,
        title: Any,
        items: Union[List[Tuple[Any, ...]], List[str]],
        dropselect_id: str = '',
        default: Optional[int] = None,
        onchange: CallbackType = None,
        onreturn: CallbackType = None,
        onselect: CallbackType = None,
        open_middle: bool = False,
        placeholder: str = 'Select an option',
        placeholder_add_to_selection_box: bool = True,
        scrollbar_color: ColorInputType = (235, 235, 235),
        scrollbar_cursor: CursorInputType = None,
        scrollbar_shadow: bool = False,
        scrollbar_shadow_color: ColorInputType = (0, 0, 0),
        scrollbar_shadow_offset: int = 2,
        scrollbar_shadow_position: str = POSITION_NORTHWEST,
        scrollbar_slider_color: ColorInputType = (200, 200, 200),
        scrollbar_slider_hover_color: ColorInputType = (170, 170, 170),
        scrollbar_slider_pad: NumberType = 0,
        scrollbar_thick: int = 20,
        scrollbars: str = POSITION_SOUTHEAST,
        selection_box_arrow_color: ColorInputType = (150, 150, 150),
        selection_box_arrow_margin: Tuple3IntType = (5, 5, 0),
        selection_box_bgcolor: ColorInputType = (255, 255, 255),
        selection_box_border_color: ColorInputType = (150, 150, 150),
        selection_box_border_width: int = 1,
        selection_box_height: int = 3,
        selection_box_inflate: Tuple2IntType = (0, 0),
        selection_box_margin: Tuple2NumberType = (25, 0),
        selection_box_text_margin: int = 5,
        selection_box_width: int = 0,
        selection_infinite: bool = False,
        selection_option_border_color: ColorInputType = (220, 220, 220),
        selection_option_border_width: int = 1,
        selection_option_cursor: CursorInputType = None,
        selection_option_font: Optional[FontType] = None,
        selection_option_font_color: ColorInputType = (0, 0, 0),
        selection_option_font_size: Optional[int] = None,
        selection_option_padding: PaddingType = 5,
        selection_option_selected_bgcolor: ColorInputType = (188, 227, 244),
        selection_option_selected_font_color: ColorInputType = (0, 0, 0),
        *args,
        **kwargs
    ) -> None:
        assert isinstance(default, (int, type(None)))
        assert isinstance(dropselect_id, str)
        assert isinstance(items, list)
        assert isinstance(open_middle, bool)
        assert isinstance(placeholder, str)
        assert isinstance(placeholder_add_to_selection_box, bool)

        # Check items list
        check_selector_items(items)
        if default is None:
            default = -1
        assert default >= -1, \
            'default position must be equal or greater than zero'
        assert default < len(items), \
            'default position should be lower than number of values'
        assert isinstance(dropselect_id, str), 'id must be a string'
        assert isinstance(default, int), 'default must be an integer'

        # Check styling
        assert isinstance(scrollbar_shadow, bool)
        assert isinstance(scrollbar_shadow_offset, int) and scrollbar_shadow_offset >= 0
        assert isinstance(scrollbar_slider_pad, NumberInstance) and scrollbar_slider_pad >= 0
        assert isinstance(scrollbar_thick, int) and scrollbar_thick >= 0
        assert isinstance(selection_box_border_width, int) and selection_box_border_width >= 0
        assert isinstance(selection_box_height, int) and selection_box_height >= 1
        assert isinstance(selection_box_text_margin, int) and selection_box_text_margin >= 0
        assert isinstance(selection_box_width, int) and selection_box_width >= 0
        assert isinstance(selection_infinite, bool)
        assert isinstance(selection_option_border_width, int) and selection_option_border_width >= 0
        assert isinstance(selection_option_padding, PaddingInstance)
        assert_cursor(scrollbar_cursor)
        assert_cursor(selection_option_cursor)
        assert_position(scrollbar_shadow_position)
        assert_position(scrollbars)
        assert_vector(selection_box_arrow_margin, 3, int)
        assert_vector(selection_box_inflate, 2, int)
        assert_vector(selection_box_margin, 2)
        scrollbar_color = assert_color(scrollbar_color)
        scrollbar_shadow_color = assert_color(scrollbar_shadow_color)
        scrollbar_slider_color = assert_color(scrollbar_slider_color)
        scrollbar_slider_hover_color = assert_color(scrollbar_slider_hover_color)
        selection_box_arrow_color = assert_color(selection_box_arrow_color)
        selection_box_bgcolor = assert_color(selection_box_bgcolor)
        selection_box_border_color = assert_color(selection_box_border_color)
        selection_option_border_color = assert_color(selection_option_border_color)
        selection_option_font_color = assert_color(selection_option_font_color)
        selection_option_selected_bgcolor = assert_color(selection_option_selected_bgcolor)
        selection_option_selected_font_color = assert_color(selection_option_selected_font_color)

        if selection_option_font is not None:
            assert_font(selection_option_font)
        if selection_option_font_size is not None:
            assert isinstance(selection_option_font_size, int) and selection_option_font_size > 0

        super(DropSelect, self).__init__(
            onchange=onchange,
            onreturn=onreturn,
            onselect=onselect,
            title=title,
            widget_id=dropselect_id,
            args=args,
            kwargs=kwargs
        )

        self._accept_events = True
        self._close_on_apply = True
        self._default_value = default
        self._drop_frame = None
        self._index = default
        self._items = items.copy()
        self._open_bottom = True
        self._open_middle = open_middle
        self._placeholder = placeholder
        self._placeholder_add_to_selection_box = placeholder_add_to_selection_box
        self._selection_effect_draw_post = False
        self._title_size = (0, 0)

        # If True adds a space equals to the height of the option at left, used for
        # drawing some options (for example, ticks, boxes, etc)
        self._selection_option_left_space = False
        self._selection_option_left_space_height_factor = 1
        self._selection_option_left_space_margin = (0, 0, 0)  # left, right, top

        # Style
        self._option_font = None
        self._scrollbar_color = scrollbar_color
        self._scrollbar_cursor = scrollbar_cursor
        self._scrollbar_shadow = scrollbar_shadow
        self._scrollbar_shadow_color = scrollbar_shadow_color
        self._scrollbar_shadow_offset = scrollbar_shadow_offset
        self._scrollbar_shadow_position = scrollbar_shadow_position
        self._scrollbar_slider_color = scrollbar_slider_color
        self._scrollbar_slider_hover_color = scrollbar_slider_hover_color
        self._scrollbar_slider_pad = scrollbar_slider_pad
        self._scrollbar_thick = scrollbar_thick
        self._scrollbars = scrollbars
        self._selection_box_arrow_color = selection_box_arrow_color
        self._selection_box_arrow_margin = selection_box_arrow_margin
        self._selection_box_bgcolor = selection_box_bgcolor
        self._selection_box_border_color = selection_box_border_color
        self._selection_box_border_width = selection_box_border_width
        self._selection_box_height = selection_box_height
        self._selection_box_inflate = selection_box_inflate
        self._selection_box_margin = (int(selection_box_margin[0]), int(selection_box_margin[1]))
        self._selection_box_text_margin = selection_box_text_margin
        self._selection_box_width = selection_box_width
        self._selection_infinite = selection_infinite
        self._selection_option_border_color = selection_option_border_color
        self._selection_option_border_width = selection_option_border_width
        self._selection_option_cursor = selection_option_cursor
        self._selection_option_padding = parse_padding(selection_option_padding)
        self._selection_option_selected_bgcolor = selection_option_selected_bgcolor

        self._selection_option_font_style = {
            'color': selection_option_font_color,
            'color_selected': selection_option_selected_font_color,
            'name': selection_option_font,
            'size': selection_option_font_size
        }

        self.active = False

    def set_default_value(self, index: int) -> 'DropSelect':
        self._default_value = index
        return self

    def _apply_font(self) -> None:
        # Compute title size
        title_render = self._font.size(self._title)
        w = int(title_render[0] + self._selection_box_margin[0] - self._selection_box_inflate[0] / 2)
        h = int(title_render[1] + self._selection_box_inflate[1] / 2 - self._selection_box_border_width)
        self._title_size = (w, h)

        # Load option font
        if self._selection_option_font_style['name'] is None:
            self._selection_option_font_style['name'] = self._font_name
        if self._selection_option_font_style['size'] is None:
            self._selection_option_font_style['size'] = int(self._font_size)
        self._option_font = get_font(self._selection_option_font_style['name'], self._selection_option_font_style['size'])
        if self._selection_box_width == 0:
            f = self._render_option_string(self._placeholder)
            h = self._render_string(self._title, self.get_font_color_status()).get_height()
            self._selection_box_width = int(f.get_width()
                                            + self._selection_box_arrow_margin[0]
                                            + self._selection_box_arrow_margin[1]
                                            + h - h / 4
                                            + 2 * self._selection_box_border_width)

    def _make_selection_drop(self) -> 'DropSelect':
        """
        Make the selection drop box. This function creates the box UI and adds
        the selection options. Should be called after the widget is configured.

        :return: Self reference
        """
        # noinspection PyUnresolvedReferences
        from pygame_menu._scrollarea import get_scrollbars_from_position

        if not self.configured:
            raise RuntimeError(f'{self.get_class_id()} must be configured before creating selection drop')

        # Create options buttons
        total_height = 0
        max_height = 0
        frame_width = self._selection_box_width + self._selection_box_inflate[0]
        self._option_buttons = []

        # Add placeholder button
        if self._placeholder_add_to_selection_box:
            self._items.insert(0, (self._placeholder, -1))

        for opt_id in range(len(self._items)):
            option = self._items[opt_id]
            btn = Button(option[0],
                         onreturn=self._click_option,
                         index=opt_id - (1 if self._placeholder_add_to_selection_box else 0),
                         button_id=self._id + '+option-' + uuid4(short=True))
            btn.set_background_color(
                color=self._selection_box_bgcolor
            )
            btn.set_border(
                width=self._selection_option_border_width,
                color=self._selection_option_border_color
            )
            btn.set_controls(
                joystick=False,  # Only drop select controls the joystick behaviour
                mouse=self._mouse_enabled,
                touchscreen=self._touchscreen_enabled,
                keyboard=False  # Only drop select controls the keyboard behaviour
            )
            btn.set_cursor(  # This feature does not work properly
                cursor=self._selection_option_cursor
            )
            font_c = self._selection_option_font_style['color']
            if self._placeholder_add_to_selection_box:
                font_color = font_c if opt_id != 0 else self._font_readonly_color
            else:
                font_color = font_c
            btn.set_font(
                antialias=self._font_antialias,
                background_color=None,
                color=font_color,
                font=self._selection_option_font_style['name'],
                font_size=self._selection_option_font_style['size'],
                readonly_color=self._font_readonly_color,
                readonly_selected_color=self._font_readonly_selected_color,
                selected_color=self._font_selected_color
            )
            btn.set_padding(
                padding=self._selection_option_padding
            )
            btn.add_self_to_kwargs('btn')
            btn.set_tab_size(self._tab_size)
            btn.configured = True
            btn.set_menu(self._menu)
            btn._update__repr___(self)

            self._option_buttons.append(btn)

            bh = btn.get_height() - self._selection_option_border_width
            if self._selection_option_left_space and not (self._placeholder_add_to_selection_box and opt_id == 0):
                prev_pad = btn._padding  # top, right, bottom, left
                prev_pad_t: Tuple4IntType = btn._padding_transform
                dh = int(btn.get_height(apply_padding=False) * self._selection_option_left_space_height_factor)
                btn.set_attribute('left_space_height', dh)
                m = self._selection_option_left_space_margin
                btn._padding = prev_pad[0], prev_pad[1], prev_pad[2], prev_pad[3] + dh + m[0] + m[1]
                btn._padding_transform = prev_pad_t[0], prev_pad_t[1], prev_pad_t[2], prev_pad_t[3] + dh + m[0] + m[1]

            total_height += bh
            if opt_id + 1 <= self._selection_box_height:
                max_height += bh

        max_width = frame_width
        if total_height != max_height:
            max_width -= self._scrollbar_thick
            frame_width -= self._scrollbar_thick
            total_height -= self._selection_box_border_width
            max_height -= self._selection_box_border_width
        elif total_height > 0:
            total_height += self._selection_box_border_width
            max_height += self._selection_box_border_width
        for btn in self._option_buttons:
            max_width = max(max_width, btn.get_width() - self._selection_option_border_width)

        # Update options rect delta width
        for btn in self._option_buttons:
            btn._rect_size_delta = (max_width - btn.get_width(), 0)

        # Pop placeholder
        if self._placeholder_add_to_selection_box:
            self._items.pop(0)
            placeholder_button = self._option_buttons.pop(0)
        else:
            placeholder_button = None

        # Unpack previous frame (if exists)
        if self._drop_frame is not None:
            self._drop_frame.set_menu(None)

        # Create frame
        self._drop_frame = Frame(max_width, max(total_height, 1), ORIENTATION_VERTICAL, frame_id=self._id + '+frame-' + uuid4(short=True))
        self._drop_frame._accepts_title = False
        self._drop_frame._menu_can_be_none_pack = True
        self._drop_frame.hide()
        self._drop_frame.set_background_color(
            color=self._selection_box_bgcolor
        )
        self._drop_frame.set_border(
            width=self._selection_box_border_width,
            color=self._selection_box_border_color
        )
        self._drop_frame.set_scrollarea(self._scrollarea)
        self._drop_frame.relax()
        self._drop_frame.configured = True
        self._drop_frame.set_tab_size(self._tab_size)
        self._drop_frame._update__repr___(self)
        self._drop_frame.set_controls(
            joystick=self._joystick_enabled,
            mouse=self._mouse_enabled,
            touchscreen=self._touchscreen_enabled,
            keyboard=self._keyboard_enabled
        )

        if total_height > 0:
            # Menu is needed while creating the scrollarea, as that reference
            # is later passed to the scrollbars, which need menu for touchscreen
            # events
            self._drop_frame._menu = self._menu
            self._drop_frame.make_scrollarea(
                max_width=frame_width,
                max_height=max_height,
                scrollarea_color=self._selection_box_arrow_color,
                scrollbar_color=self._scrollbar_color,
                scrollbar_cursor=self._scrollbar_cursor,
                scrollbar_shadow_color=self._scrollbar_shadow_color,
                scrollbar_shadow_offset=self._scrollbar_shadow_offset,
                scrollbar_shadow_position=self._scrollbar_shadow_position,
                scrollbar_shadow=self._scrollbar_shadow,
                scrollbar_slider_color=self._scrollbar_slider_color,
                scrollbar_slider_hover_color=self._scrollbar_slider_hover_color,
                scrollbar_slider_pad=self._scrollbar_slider_pad,
                scrollbar_thick=self._scrollbar_thick,
                scrollbars=get_scrollbars_from_position(self._scrollbars)
            )

        # Menu is formerly set in frame. That method adds the frame to update frames
        # depending on if frame is scrollable or not
        self._drop_frame.set_menu(self._menu)

        self._drop_frame.set_scrollarea(self._scrollarea)
        if self._frame is not None:
            self._drop_frame.set_frame(self._frame)

        # Set sizing properties
        if total_height > 0:
            add_scrollbar = self._scrollbar_thick if max_width != frame_width else 0
            border_w = self._selection_box_border_width if total_height != max_height else 0
            self._drop_frame.set_attribute('height', max_height + add_scrollbar - border_w)
            self._drop_frame.set_attribute('width', frame_width)
        else:
            self._drop_frame.set_attribute('height', 0)
            self._drop_frame.set_attribute('width', 0)
            if self._placeholder_add_to_selection_box:
                placeholder_button.hide()
        margin_width = self._selection_box_border_width if total_height == max_height else 0
        self._drop_frame.set_attribute('extra_margin', margin_width)
        self._drop_frame.set_attribute('placeholder_button', placeholder_button)

        # Pack options
        if self._placeholder_add_to_selection_box:
            self._drop_frame.pack(placeholder_button)
        for opt in self._option_buttons:
            self._drop_frame.pack(opt, margin=(0, -self._selection_option_border_width))

        # Update options if index is defined
        if self._index != -1:
            self.set_value(self._index)

        return self

    def _append_to_menu(self) -> None:
        self._make_selection_drop()

    def on_remove_from_menu(self) -> 'DropSelect':
        if self._drop_frame is not None:
            self._drop_frame.set_menu(None)
        return self

    def hide(self) -> 'DropSelect':
        super(DropSelect, self).hide()
        if self._drop_frame is not None:
            self._drop_frame.hide()
        return self

    def show(self) -> 'DropSelect':
        super(DropSelect, self).show()
        if self.active:
            self._toggle_drop()
        return self

    def scrollh(self, value: NumberType) -> 'DropSelect':
        """
        Scroll drop frame to horizontal value.

        :param value: Horizontal scroll value, if ``0`` scroll to left; ``1`` scroll to right
        :return: Self reference
        """
        if self._drop_frame is not None:
            self._drop_frame.scrollh(value)
        return self

    def scrollv(self, value: NumberType) -> 'DropSelect':
        """
        Scroll drop frame to vertical value.

        :param value: Vertical scroll value, if ``0`` scroll to top; ``1`` scroll to bottom
        :return: Self reference
        """
        if self._drop_frame is not None:
            self._drop_frame.scrollv(value)
        return self

    def get_scroll_value_percentage(self, orientation: str) -> float:
        """
        Get the scroll value in percentage, if ``0`` the scroll is at top/left,
        ``1`` bottom/right.

        .. note::

            If ScrollArea does not contain such orientation scroll, or frame is
            not scrollable, ``-1`` is returned.

        :param orientation: Orientation. See :py:mod:`pygame_menu.locals`
        :return: Value from ``0`` to ``1``
        """
        if self._drop_frame is not None:
            return self._drop_frame.get_scroll_value_percentage(orientation)
        return -1

    def set_scrollarea(self, scrollarea: 'pygame_menu._scrollarea.ScrollArea') -> None:
        super(DropSelect, self).set_scrollarea(scrollarea)
        if self._drop_frame is not None:
            self._drop_frame.set_scrollarea(scrollarea)

    def set_frame(self, frame: 'pygame_menu.widgets.Frame') -> 'DropSelect':
        super(DropSelect, self).set_frame(frame)
        if self._drop_frame is not None:
            self._drop_frame.set_frame(frame)
        return self

    def _click_option(self, index: int, btn: 'Button') -> None:
        """
        Function triggered after option has been selected or clicked.

        :param index: Option index within list
        """
        btn.set_attribute('ignore_scroll_to_widget')
        prev_index = self._index
        self.set_value(index)
        if self._index != prev_index and self._index != -1:
            self.change(*self._items[self._index][1:])
        if self._close_on_apply:
            self.active = False
            if self._drop_frame is not None:
                self._drop_frame.hide()
        btn.remove_attribute('ignore_scroll_to_widget')

    def set_position(self, x: NumberType, y: NumberType) -> 'DropSelect':
        super(DropSelect, self).set_position(x, y)
        if self._drop_frame is not None:
            x = self._rect.x
            y = self._rect.y
            if not self._open_middle:
                if self._open_bottom:
                    self._drop_frame.set_position(x + self._title_size[0],
                                                  y + self._title_size[1]
                                                  + self.get_attribute('delta_title_height', 0))
                else:
                    self._drop_frame.set_position(x + self._title_size[0],
                                                  y - self._drop_frame.get_attribute('height')
                                                  + self._drop_frame.get_attribute('extra_margin'))
            else:
                self._drop_frame.set_position(*self._compute_position_middle())
            for w in self._option_buttons:
                w._set_position_relative_to_frame()
            if self._placeholder_add_to_selection_box:
                placeholder_button: 'Button' = self._drop_frame.get_attribute('placeholder_button')
                placeholder_button._set_position_relative_to_frame()
            self._drop_frame.update_position()
        return self

    def scale(self, *args, **kwargs) -> 'DropSelect':
        raise WidgetTransformationNotImplemented()

    def resize(self, *args, **kwargs) -> 'DropSelect':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'DropSelect':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'DropSelect':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'DropSelect':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'DropSelect':
        raise WidgetTransformationNotImplemented()

    def _draw(self, surface: 'pygame.Surface') -> None:
        surface.blit(self._surface, self._rect.topleft)

    def draw_after_if_selected(self, surface: Optional['pygame.Surface']) -> 'DropSelect':
        super(DropSelect, self).draw_after_if_selected(surface)
        if self.active and self.is_visible():
            self._check_drop_made()

            if self._open_middle:
                # Draw drop frame in menu widgets surface
                assert self._menu is not None, 'middle position need menu reference'
                surface = self._menu._widgets_surface

            self._drop_frame.draw(surface)
            self.last_surface = surface

        return self

    def _render_option_string(self, text: str) -> 'pygame.Surface':
        """
        Render option string surface.

        :param text: Text to render
        :return: Option string surface
        """
        color = self._selection_option_font_style['color']
        if self.readonly or self._index == -1:
            color = self._font_readonly_color
        text = text.replace('\t', ' ' * self._tab_size)
        return self._option_font.render(text, self._font_antialias, color)

    def _get_current_selected_text(self) -> str:
        """
        Return the current selected text.

        :return: Text
        """
        if self._index == -1:
            current_selected = self._placeholder
        else:
            current_selected = self.get_value()[0][0]
        return current_selected

    def _render(self) -> Optional[bool]:
        if self._option_font is None:
            return

        scroll_v = 0
        menu_height = 0 if self._menu is None else self._menu.get_height(widget=True)
        current_selected = self._get_current_selected_text()

        if not self._render_hash_changed(current_selected, self._selected, self._visible, self._index,
                                         self.readonly, self.active, self._open_bottom, scroll_v,
                                         menu_height, self._open_middle, len(self._items), self._rect.x, self._rect.y):
            return True

        title = self._render_string(self._title, self.get_font_color_status())
        current = self._render_option_string(current_selected)

        # Compute virtual inflate to be applied to current rect
        vi = 0  # Virtual inflate
        if current.get_height() < title.get_height():
            vi = (title.get_height() - current.get_height()) / 2

        current_rect_bg = current.get_rect()
        current_rect_bg.x += title.get_width() + self._selection_box_margin[0]
        current_rect_bg.y += int((self._selection_box_inflate[1]) / 2 + vi + self._selection_box_margin[1])
        current_rect_bg.width = self._selection_box_width
        current_rect_bg = current_rect_bg.inflate(
            (int(self._selection_box_inflate[0]), int(self._selection_box_inflate[1] + 2 * vi))
        )

        # Compute delta title if height is lower than selection box
        h = title.get_height()
        delta_title_height = max(int(math.floor((current_rect_bg.height - h) / 2)), 0)

        # Create arrows
        arrow = pygame.Rect(
            int(title.get_width()
                + self._selection_box_margin[0]
                + self._selection_box_width - h),
            int(self._selection_box_arrow_margin[2]
                + (self._selection_box_inflate[1] + vi) / 4
                + delta_title_height
                + self._selection_box_margin[1]),
            h,
            h
        )
        w = h + self._selection_box_arrow_margin[1]

        # Check which direction it should open (top, bottom)
        self._open_bottom = True
        if self._drop_frame is not None and self._scrollarea is not None:
            assert self.get_scrollarea().get_world() is not None, \
                'scrollarea world is None. The widget have been appended to the ' \
                'Menu?'
            rect = self._rect.copy()
            rect.width = self._selection_box_width
            rect.y += delta_title_height
            rect.x += self._title_size[0]
            rect.height += self._drop_frame.get_attribute('height')
            if rect.width != 0 and rect.height != 0:
                rect_clipped = self._scrollarea.get_world_rect().clip(rect)
                if rect.height != rect_clipped.height:
                    self._open_bottom = False

        # Create frame
        if self._drop_frame is not None and self._open_middle and self._menu is not None:
            self._drop_frame.set_scrollarea(self._menu.get_scrollarea())
            self._drop_frame._frame = None
            if not self._menu._mouse_motion_selection:
                self.force_menu_draw_focus = True
        else:
            self.force_menu_draw_focus = False

        delta_title_height_dh = math.ceil((current_rect_bg.height - h) / 2) + 2 * self._selection_box_border_width
        if delta_title_height != 0:
            self.set_attribute('delta_title_height', delta_title_height_dh)

        arrow_up = (
            (int(arrow.right - w + h / 2 - h / 16), int(arrow.centery - h / 6 - h / 20)),
            (int(arrow.right - w + h / 2 + h / 4 - h / 16), int(arrow.centery + h / 4 - h / 20)),
            (int(arrow.right - w + h - h / 16), int(arrow.centery - h / 6 - h / 20))
        )
        arrow_down = (
            (int(arrow.right - w + h / 2 - h / 16), int(arrow.centery + h / 4 - h / 20)),
            (int(arrow.right - w + h / 2 + h / 4 - h / 16), int(arrow.centery - h / 6 - h / 20)),
            (int(arrow.right - w + h - h / 16), int(arrow.centery + h / 4 - h / 20))
        )
        if not self._open_bottom:
            if not self.active:
                arrow_right_pos = arrow_down
            else:
                arrow_right_pos = arrow_up
        else:
            if not self.active:
                arrow_right_pos = arrow_up
            else:
                arrow_right_pos = arrow_down

        self._surface = make_surface(title.get_width()
                                     + self._selection_box_margin[0]
                                     + self._selection_box_width
                                     + self._selection_box_inflate[0] / 2
                                     + self._selection_box_border_width,
                                     max(title.get_height() + self._selection_box_inflate[1],
                                         current_rect_bg.height - self.active * self._selection_box_border_width))
        self._surface.blit(title, (0, int(self._selection_box_inflate[1] / 2 + delta_title_height)))
        pygame.draw.rect(self._surface, self._selection_box_bgcolor, current_rect_bg)
        pygame.draw.rect(self._surface, self._selection_box_border_color, current_rect_bg,
                         self._selection_box_border_width)

        # Crop current max width
        cropped_current_w = self._selection_box_width \
                            - self._selection_box_arrow_margin[0] \
                            - self._selection_box_arrow_margin[1] \
                            - h / 2 \
                            - h / 16 \
                            - self._selection_box_text_margin
        assert cropped_current_w > 0, \
            'there is no left space for text width, try increasing selection_box_width size'
        new_current = make_surface(cropped_current_w, current.get_height())
        new_current.blit(current, (0, 0))
        # new_current.fill((0, 0, 0))
        self._surface.blit(new_current,
                           (
                               int(title.get_width()
                                   + self._selection_box_margin[0]
                                   + self._selection_box_text_margin),
                               int(self._selection_box_inflate[1] / 2
                                   + vi - 1
                                   + self._selection_box_margin[1])
                           ))
        if len(self._items) > 0:
            pygame.draw.polygon(self._surface, self._selection_box_arrow_color, arrow_right_pos)

        self._rect.width, self._rect.height = self._surface.get_size()
        self.force_menu_surface_update()

    def get_index(self) -> int:
        """
        Get selected index.

        :return: Selected index
        """
        return self._index

    def get_value(self) -> Tuple[Union[Tuple[Any, ...], str], int]:
        """
        Return the current value of the selected index.

        :return: Value and index as a tuple, (value, index)
        """
        if self._index == -1:
            raise ValueError('any item has been selected yet as index is -1')
        return self._items[self._index], self._index

    def value_changed(self) -> bool:
        return self._index != self._default_value

    def _down(self) -> None:
        """
        Move current selection down.
        """
        if self.readonly:
            return
        if len(self._items) == 0:
            return
        if not self.active:
            return self._toggle_drop()
        if self._index == -1:
            self.set_value(len(self._items) - 1)
        else:
            if self._selection_infinite:
                self.set_value((self._index - 1) % len(self._items))
            else:
                prev = self._index
                new = max(0, self._index - 1)
                if prev == new:
                    return
                self.set_value(new)
        self.change(*self._items[self._index][1:])
        self._sound.play_key_add()

    def _up(self) -> None:
        """
        Move current selection up.
        """
        if self.readonly:
            return
        if len(self._items) == 0:
            return
        if not self.active:
            return self._toggle_drop()
        if self._index == -1:
            self.set_value(0)
        else:
            if self._selection_infinite:
                self.set_value((self._index + 1) % len(self._items))
            else:
                prev = self._index
                new = min(self._index + 1, len(self._items) - 1)
                if prev == new:
                    return
                self.set_value(new)
        self.change(*self._items[self._index][1:])
        self._sound.play_key_add()

    def set_value(self, item: Union[str, int]) -> None:
        """
        Set the current value of the widget, selecting the item that matches the
        text if ``item`` is a string, or the index if ``item`` is an integer.
        This method raises ``ValueError`` if no item found.

        For example, if widget item list is ``[['a', 0], ['b', 1], ['a', 2]]``:

        - *widget*.set_value('a') -> Widget selects the first item (index 0)
        - *widget*.set_value(2) -> Widget selects the third item (index 2)

        .. note::

            This method does not trigger any event (change).

        :param item: Item to select, can be a string or an integer
        """
        assert isinstance(item, (str, int)), 'item must be a string or an integer'

        if isinstance(item, str):
            found = False
            for i in self._items:
                if i[0] == item:
                    self._index = self._items.index(i)
                    found = True
                    break
            if not found:
                raise ValueError(f'no value "{item}" found in drop select')
        elif isinstance(item, int):
            assert -1 <= item < len(self._items), \
                'item index must be greater than zero and lower than the number ' \
                'of items on the drop select'
            self._index = item

        # Update options background selection
        for b_ind_x in range(len(self._option_buttons)):
            btn = self._option_buttons[b_ind_x]
            if b_ind_x == self._index:
                btn.set_background_color(self._selection_option_selected_bgcolor)
                btn.update_font({
                    'color': self._selection_option_font_style['color_selected']
                })
                if not self._drop_frame.has_attribute('ignorescroll'):
                    btn.scroll_to_widget(scroll_parent=False)
            else:
                btn.set_background_color(self._selection_box_bgcolor)
                btn.update_font({'color': self._selection_option_font_style['color']})

        # Force render
        self._render()

    def update_items(self, items: Union[List[Tuple[Any, ...]], List[str]]) -> None:
        """
        Update drop select items.

        .. note::

            If the length of the list is different from the previous one, the new
            index of the select will be the ``-1``, that is, same as the unselected
            status.

        :param items: New drop select items; format ``[('Item1', a, b, c...), ('Item2', d, e, f...)]``
        """
        assert isinstance(items, list)
        if len(items) > 0:
            check_selector_items(items)
        if self._index != -1 and len(items) != len(self._items) or (self._index + 1) > len(items):
            self._index = -1
            self._default_value = -1
        self._items = items.copy()
        if self._drop_frame is not None:
            self._drop_frame.set_menu(None)
        self._drop_frame = None
        self.active = False
        self._make_selection_drop()

    def _check_drop_made(self) -> None:
        """
        Checks if drop selection has been made.
        """
        if self._drop_frame is None:
            raise _SelectionDropNotMadeException(
                f'selection drop has not been made yet. Call {self.get_class_id()}'
                f'._make_selection_drop() for avoiding this exception'
            )

    def get_items(self) -> Union[List[Tuple[Any, ...]], List[str]]:
        """
        Return a copy of the select items.

        :return: Select items list
        """
        return self._items.copy()

    def _toggle_drop(self) -> None:
        """
        Open drop selection.
        """
        self._check_drop_made()
        if not self._selected:
            return
        if len(self._items) == 0:
            return
        self.active = not self.active
        if self.active:
            if self._drop_frame is not None:
                self._drop_frame.show()
            self.scroll_to_widget(scroll_parent=False)
        else:
            if self._drop_frame is not None:
                self._drop_frame.hide()
        if self.active and self._index != -1:
            self.set_value(self._index)

    def _compute_position_middle(self, add_offset: bool = True) -> Tuple2IntType:
        """
        Compute box position if position is in the middle.

        :param add_offset: Adds offset
        :return: Position
        """
        assert self._menu is not None, \
            'menu cannot be none if the position is in middle (open_middle)'
        self._check_drop_made()
        if add_offset:
            offx, offy = self._menu.get_scrollarea().get_offsets()
        else:
            offx, offy = 0, 0
        w, h = self._menu.get_size()
        h -= self._menu.get_menubar().get_height()
        bw, bh = self._selection_box_width, self._drop_frame.get_attribute('height')
        assert w >= bw, \
            f'selection box width ({bw}) cannot be greater than menu width ({w})'
        assert h >= bh, \
            f'selection box height ({bh}) cannot be greater than menu height ({h})'
        x = (w - bw) / 2 + offx
        y = (h - bh) / 2 + offy
        return x, y

    def get_focus_rect(self) -> 'pygame.Rect':
        self._check_drop_made()
        rect = self.get_rect(apply_padding=False, to_real_position=True)
        if self.active:
            rect.width = self._selection_box_width
            rect.x += self._title_size[0]
            rect.height += self._drop_frame.get_attribute('height') - self._drop_frame.get_attribute('extra_margin')
            if not self._open_bottom:
                rect.y -= self._drop_frame.get_attribute('height') - self._drop_frame.get_attribute('extra_margin')
            if self._open_middle:
                x, y = self._compute_position_middle(add_offset=False)
                rect.x = int(x + self._menu.get_position()[0])
                rect.y = int(y + self._menu.get_scrollarea().get_position()[1])
                rect.height -= self._rect.height
        else:
            rect.width -= self._selection_box_border_width
        return rect

    def _drop_frame_scrolling(self) -> bool:
        """
        Checks if drop frame is scrolling.

        :return: ``True`` if scrolling
        """
        if self._drop_frame is None or self._drop_frame.get_scrollarea(inner=True) is None:
            return False
        return self._drop_frame.get_scrollarea(inner=True).is_scrolling()

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)

        if self.readonly or not self.is_visible():
            self._readonly_check_mouseover(events)
            return False

        # Check scroll
        self._check_drop_made()
        updated = self._drop_frame.update(events)
        if updated:
            return True

        for event in events:

            if event.type == pygame.KEYDOWN:  # Check key is valid
                if self._ignores_keyboard_nonphysical() and not check_key_pressed_valid(event):
                    continue

            # Check mouse over
            self._check_mouseover(event)

            # Events
            keydown = self._keyboard_enabled and event.type == pygame.KEYDOWN
            joy_hatmotion = self._joystick_enabled and event.type == pygame.JOYHATMOTION
            joy_axismotion = self._joystick_enabled and event.type == pygame.JOYAXISMOTION
            joy_button_down = self._joystick_enabled and event.type == pygame.JOYBUTTONDOWN

            # Left button
            if (
                keydown and self._ctrl.move_down(event, self) or
                joy_hatmotion and self._ctrl.joy_left(event, self) or
                joy_axismotion and self._ctrl.joy_axis_x_left(event, self)
            ):
                if not self.active:
                    continue
                self._down()
                return True

            # Right button
            elif (
                keydown and self._ctrl.move_up(event, self) or
                joy_hatmotion and self._ctrl.joy_right(event, self) or
                joy_axismotion and self._ctrl.joy_axis_x_right(event, self)
            ):
                if not self.active:
                    continue
                self._up()
                return True

            # Press enter
            elif keydown and self._ctrl.apply(event, self) or joy_button_down and self._ctrl.joy_select(event, self):
                if self.active and self._index >= 0:
                    self._sound.play_key_add()
                    self.apply(*self._items[self._index][1:])
                if not (self.active and (not self._close_on_apply and self._index != -1)):
                    self._toggle_drop()
                return True

            # Press keys which active the drop but not apply
            elif keydown and self._ctrl.tab(event, self):
                self._toggle_drop()
                return True

            # Close the selection
            elif keydown and (self._ctrl.escape(event, self) or self._ctrl.back(event, self)):
                if self.active:
                    self._toggle_drop()
                return True

            # Click on dropselect; don't consider the mouse wheel (button 4 & 5)
            elif self.active and (
                event.type == pygame.MOUSEBUTTONDOWN and self._mouse_enabled and event.button in (1, 2, 3) or
                event.type == FINGERDOWN and self._touchscreen_enabled and self._menu is not None and not self._drop_frame_scrolling()
            ):
                event_pos = get_finger_pos(self._menu, event)
                if self._drop_frame.get_rect(apply_padding=False, to_real_position=True).collidepoint(*event_pos):
                    return True

            # Click on dropselect; don't consider the mouse wheel (button 4 & 5)
            elif (
                event.type == pygame.MOUSEBUTTONUP and self._mouse_enabled and event.button in (1, 2, 3) or
                event.type == FINGERUP and self._touchscreen_enabled and self._menu is not None and not self._drop_frame_scrolling()
            ):
                # Check for mouse clicks within
                if self.active:
                    for btn in self._option_buttons:
                        btn.set_attribute('ignore_scroll_to_widget')
                        updated = btn.update([event])
                        try:
                            btn.remove_attribute('ignore_scroll_to_widget')
                        except IndexError:
                            pass
                        if updated:
                            return True

                # Get event position based on input type
                event_pos = get_finger_pos(self._menu, event)

                # If collides
                rect = self.get_rect(to_real_position=True, apply_padding=False)
                if rect.collidepoint(*event_pos):
                    # Check if mouse collides left or right as percentage, use only X coordinate
                    mouse_x, _ = event_pos
                    topleft, _ = rect.topleft
                    topright, _ = rect.topright
                    dist = mouse_x - (topleft + self._title_size[0])  # Distance from title
                    if dist > 0:  # User clicked the options, not title
                        self._toggle_drop()
                        return True

                else:
                    if self.active and not self.get_focus_rect().collidepoint(*event_pos):
                        self._toggle_drop()
                        return True

            # Check mousemove
            # elif event.type == pygame.MOUSEMOTION:
            #     for btn in self._option_buttons:
            #         btn._check_mouseover(event)

        return False


class _SelectionDropNotMadeException(Exception):
    """
    Exception thrown if drop selection has not been made.
    """
    pass


class DropSelectManager(AbstractWidgetManager, ABC):
    """
    DropSelect manager.
    """

    def dropselect(
        self,
        title: Any,
        items: Union[List[Tuple[Any, ...]], List[str]],
        default: Optional[int] = None,
        dropselect_id: str = '',
        onchange: CallbackType = None,
        onreturn: CallbackType = None,
        onselect: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]] = None,
        open_middle: bool = False,
        placeholder: str = 'Select an option',
        placeholder_add_to_selection_box: bool = True,
        **kwargs
    ) -> 'pygame_menu.widgets.DropSelect':
        """
        Add a dropselect to the Menu: Drop select is a selector within a Frame.
        This drops a vertical frame if requested.

        Drop select can contain selectable items (options), but only one can be
        selected.

        The items of the DropSelect are:

        .. code-block:: python

            items = [('Item1', a, b, c...), ('Item2', d, e, f...)]

        The callbacks receive the current selected item, its index in the list,
        the associated arguments, and all unknown keyword arguments, where
        ``selected_item=widget.get_value()`` and ``selected_index=widget.get_index()``:

        .. code-block:: python

            onchange((selected_item, selected_index), a, b, c..., **kwargs)
            onreturn((selected_item, selected_index), a, b, c..., **kwargs)

        For example, if ``selected_index=0`` then ``selected_item=('Item1', a, b, c...)``.

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``readonly_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode
            - ``readonly_selected_color``       (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode and is selected
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``tab_size``                      (int) – Width of a tab character

        kwargs for modifying selection box/option style (Optional)
            - ``scrollbar_color``                       (tuple, list, str, int, :py:class:`pygame.Color`) – Scrollbar color
            - ``scrollbar_cursor``                      (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the scrollbars if the mouse is placed over
            - ``scrollbar_shadow_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the shadow of each scrollbar
            - ``scrollbar_shadow_offset``               (int) – Offset of the scrollbar shadow in px
            - ``scrollbar_shadow_position``             (str) – Position of the scrollbar shadow. See :py:mod:`pygame_menu.locals`
            - ``scrollbar_shadow``                      (bool) – Indicate if a shadow is drawn on each scrollbar
            - ``scrollbar_slider_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the sliders
            - ``scrollbar_slider_hover_color``          (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the slider if hovered or clicked
            - ``scrollbar_slider_pad``                  (int, float) – Space between slider and scrollbars borders in px
            - ``scrollbar_thick``                       (int) – Scrollbar thickness in px
            - ``scrollbars``                            (str) – Scrollbar position. See :py:mod:`pygame_menu.locals`
            - ``selection_box_arrow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Selection box arrow color
            - ``selection_box_arrow_margin``            (tuple) – Selection box arrow margin (left, right, vertical) in px
            - ``selection_box_bgcolor``                 (tuple, list, str, int, :py:class:`pygame.Color`) – Selection box background color
            - ``selection_box_border_color``            (tuple, list, str, int, :py:class:`pygame.Color`) – Selection box border color
            - ``selection_box_border_width``            (int) – Selection box border width
            - ``selection_box_height``                  (int) – Number of the options which can be packed before showing scroll
            - ``selection_box_inflate``                 (tuple) – Selection box inflate on x-axis and y-axis (x, y) in px
            - ``selection_box_margin``                  (tuple, list) – Selection box on x-axis and y-axis (x, y) margin from title in px
            - ``selection_box_text_margin``             (int) – Left margin of the text inside the selection box in px
            - ``selection_box_width``                   (int) – Selection box width in px. If ``0`` compute automatically to fit placeholder
            - ``selection_infinite``                    (bool) – If ``True`` selection can rotate through bottom/top
            - ``selection_option_border_color``         (tuple, list, str, int, :py:class:`pygame.Color`) – Option border color
            - ``selection_option_border_width``         (int) – Option border width
            - ``selection_option_font_color``           (tuple, list, str, int, :py:class:`pygame.Color`) – Option font color
            - ``selection_option_font_size``            (int, None) – Option font size. If ``None`` use the 75% of the widget font size
            - ``selection_option_font``                 (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Option font. If ``None`` use the same font as the widget
            - ``selection_option_padding``              (int, float, tuple, list ) – Selection padding. See padding styling
            - ``selection_option_selected_bgcolor``     (tuple, list, str, int, :py:class:`pygame.Color`) – Selected option background color
            - ``selection_option_selected_font_color``  (tuple, list, str, int, :py:class:`pygame.Color`) – Selected option font color

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Be careful with kwargs collision. Consider that all optional documented
            kwargs keys are removed from the object.

        :param title: Drop select title
        :param items: Item list of the drop select; format ``[('Item1', a, b, c...), ('Item2', d, e, f...)]``
        :param default: Index of default item to display. If ``None`` no item is selected
        :param dropselect_id: ID of the dropselect
        :param onchange: Callback when changing the drop select item
        :param onreturn: Callback when pressing return (apply) on the selected item
        :param onselect: Function when selecting the widget
        :param open_middle: If ``True`` the selection box is opened in the middle of the menu
        :param placeholder: Text shown if no option is selected yet
        :param placeholder_add_to_selection_box: If ``True`` adds the placeholder button to the selection box
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.DropSelect`
        """
        # Filter widget attributes to avoid passing them to the callbacks
        attributes = self._filter_widget_attributes(kwargs)

        # Get selection box properties
        selection_box_arrow_color = kwargs.pop('selection_box_arrow_color', self._theme.widget_box_arrow_color)
        selection_box_arrow_margin = kwargs.pop('selection_box_arrow_margin', self._theme.widget_box_arrow_margin)
        selection_box_bgcolor = kwargs.pop('selection_box_bgcolor', self._theme.widget_box_background_color)
        selection_box_border_color = kwargs.pop('selection_box_border_color', self._theme.widget_box_border_color)
        selection_box_border_width = kwargs.pop('selection_box_border_width', self._theme.widget_box_border_width)
        selection_box_height = kwargs.pop('selection_box_height', 3)
        selection_box_inflate = kwargs.pop('selection_box_inflate', self._theme.widget_border_inflate)
        selection_box_margin = kwargs.pop('selection_box_margin', self._theme.widget_box_margin)
        selection_box_text_margin = kwargs.pop('selection_box_text_margin', self._theme.widget_box_arrow_margin[0])
        selection_box_width = kwargs.pop('selection_box_width', 0)
        selection_infinite = kwargs.pop('selection_infinite', False)
        selection_option_border_color = kwargs.pop('selection_option_border_color', self._theme.scrollbar_color)
        selection_option_border_width = kwargs.pop('selection_option_border_width', self._theme.widget_box_border_width)
        # selection_option_cursor = kwargs.pop('selection_option_cursor', None)
        selection_option_font = kwargs.pop('selection_option_font', None)
        selection_option_font_color = kwargs.pop('selection_option_font_color', (0, 0, 0))
        selection_option_font_size = kwargs.pop('selection_option_font_size', None)
        selection_option_padding = kwargs.pop('selection_option_padding', (2, 5))
        selection_option_selected_bgcolor = kwargs.pop('selection_option_selected_bgcolor', (188, 227, 244))
        selection_option_selected_font_color = kwargs.pop('selection_option_selected_font_color', (0, 0, 0))

        # Get selection box scrollbar properties
        scrollbar_color = kwargs.pop('scrollbar_color', self._theme.scrollbar_color)
        scrollbar_cursor = kwargs.pop('scrollbar_cursor', self._theme.scrollbar_cursor)
        scrollbar_shadow_color = kwargs.pop('scrollbar_shadow_color', self._theme.scrollbar_shadow_color)
        scrollbar_shadow_offset = kwargs.pop('scrollbar_shadow_offset', self._theme.scrollbar_shadow_offset)
        scrollbar_shadow_position = kwargs.pop('scrollbar_shadow_position', self._theme.scrollbar_shadow_position)
        scrollbar_shadow = kwargs.pop('scrollbar_shadow', self._theme.scrollbar_shadow)
        scrollbar_slider_color = kwargs.pop('scrollbar_slider_color', self._theme.scrollbar_slider_color)
        scrollbar_slider_hover_color = kwargs.pop('scrollbar_slider_hover_color', self._theme.scrollbar_slider_hover_color)
        scrollbar_slider_pad = kwargs.pop('scrollbar_slider_pad', self._theme.scrollbar_slider_pad)
        scrollbar_thick = kwargs.pop('scrollbar_thick', self._theme.scrollbar_thick)
        scrollbars = kwargs.pop('scrollbars', self._theme.scrollarea_position)

        # Convert to tuple if enabled
        if isinstance(selection_box_margin, NumberInstance):
            selection_box_margin = (selection_box_margin, selection_box_margin)

        widget = DropSelect(
            default=default,
            dropselect_id=dropselect_id,
            items=items,
            onchange=onchange,
            onreturn=onreturn,
            onselect=onselect,
            open_middle=open_middle,
            placeholder=placeholder,
            placeholder_add_to_selection_box=placeholder_add_to_selection_box,
            scrollbar_color=scrollbar_color,
            scrollbar_cursor=scrollbar_cursor,
            scrollbar_shadow=scrollbar_shadow,
            scrollbar_shadow_color=scrollbar_shadow_color,
            scrollbar_shadow_offset=scrollbar_shadow_offset,
            scrollbar_shadow_position=scrollbar_shadow_position,
            scrollbar_slider_color=scrollbar_slider_color,
            scrollbar_slider_hover_color=scrollbar_slider_hover_color,
            scrollbar_slider_pad=scrollbar_slider_pad,
            scrollbar_thick=scrollbar_thick,
            scrollbars=scrollbars,
            selection_box_arrow_color=selection_box_arrow_color,
            selection_box_arrow_margin=selection_box_arrow_margin,
            selection_box_bgcolor=selection_box_bgcolor,
            selection_box_border_color=selection_box_border_color,
            selection_box_border_width=selection_box_border_width,
            selection_box_height=selection_box_height,
            selection_box_inflate=selection_box_inflate,
            selection_box_margin=selection_box_margin,
            selection_box_text_margin=selection_box_text_margin,
            selection_box_width=selection_box_width,
            selection_infinite=selection_infinite,
            selection_option_border_color=selection_option_border_color,
            selection_option_border_width=selection_option_border_width,
            # selection_option_cursor=selection_option_cursor,
            selection_option_font=selection_option_font,
            selection_option_font_color=selection_option_font_color,
            selection_option_font_size=selection_option_font_size,
            selection_option_padding=selection_option_padding,
            selection_option_selected_bgcolor=selection_option_selected_bgcolor,
            selection_option_selected_font_color=selection_option_selected_font_color,
            title=title,
            **kwargs
        )

        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

WIDGET
This module contains the widgets of pygame-menu.
"""

from pygame_menu.widgets.widget.button import Button
from pygame_menu.widgets.widget.colorinput import ColorInput
from pygame_menu.widgets.widget.dropselect import DropSelect
from pygame_menu.widgets.widget.dropselect_multiple import DropSelectMultiple
from pygame_menu.widgets.widget.frame import Frame
from pygame_menu.widgets.widget.hmargin import HMargin
from pygame_menu.widgets.widget.image import Image
from pygame_menu.widgets.widget.label import Label
from pygame_menu.widgets.widget.menubar import MenuBar
from pygame_menu.widgets.widget.menulink import MenuLink
from pygame_menu.widgets.widget.none import NoneWidget
from pygame_menu.widgets.widget.progressbar import ProgressBar
from pygame_menu.widgets.widget.rangeslider import RangeSlider
from pygame_menu.widgets.widget.scrollbar import ScrollBar
from pygame_menu.widgets.widget.selector import Selector
from pygame_menu.widgets.widget.surface import SurfaceWidget
from pygame_menu.widgets.widget.table import Table
from pygame_menu.widgets.widget.textinput import TextInput
from pygame_menu.widgets.widget.toggleswitch import ToggleSwitch
from pygame_menu.widgets.widget.vfill import VFill
from pygame_menu.widgets.widget.vmargin import VMargin



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLES
Example file directory.
"""

__all__ = ['create_example_window']

import pygame
import sys

from typing import Tuple

_PYGAME_ICON = [None]


# noinspection PyTypeChecker
def create_example_window(
    title: str,
    window_size: Tuple[int, int],
    pygame_menu_icon: bool = True,
    init_pygame: bool = True,
    center_window: bool = True,
    **kwargs
) -> 'pygame.Surface':
    """
    Set pygame window.

    :param title: Window title
    :param window_size: Window size
    :param pygame_menu_icon: Use pygame menu icon
    :param init_pygame: Init pygame
    :param center_window: Center the window
    :param kwargs: Optional keyword arguments received by display set_mode
    :return: Pygame surface from created display
    """
    assert len(title) > 0, 'title cannot be empty'
    assert len(window_size) == 2, 'window size shape must be (width, height)'
    assert isinstance(window_size[0], int), 'width must be an integer'
    assert isinstance(window_size[1], int), 'height must be an integer'

    from pygame_menu.baseimage import IMAGE_EXAMPLE_PYGAME_MENU, BaseImage
    import os

    if init_pygame:
        pygame.init()
    if center_window:
        os.environ['SDL_VIDEO_CENTERED'] = '1'

    # Create pygame screen and objects
    if sys.platform == 'darwin':
        kwargs = {}

    try:
        surface = pygame.display.set_mode(window_size, **kwargs)
    except TypeError:
        surface = pygame.display.set_mode(window_size)
    pygame.display.set_caption(title)

    if pygame_menu_icon:
        # noinspection PyBroadException
        try:
            if _PYGAME_ICON[0] is not None:
                pygame.display.set_icon(_PYGAME_ICON[0])
            else:
                icon = BaseImage(IMAGE_EXAMPLE_PYGAME_MENU).get_surface(new=False)
                pygame.display.set_icon(icon)
                _PYGAME_ICON[0] = icon
        except BaseException:
            pass

    return surface



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

THEMES
Theme class and predefined themes.
"""

__all__ = [

    # Main class
    'Theme',

    # Custom themes
    'THEME_BLUE',
    'THEME_DARK',
    'THEME_DEFAULT',
    'THEME_GREEN',
    'THEME_ORANGE',
    'THEME_SOLARIZED',

    # Colors
    'TRANSPARENT_COLOR'

]

import copy

import pygame_menu
from pygame_menu.baseimage import BaseImage
from pygame_menu.font import FontType, FONT_OPEN_SANS, assert_font
from pygame_menu.locals import POSITION_NORTHWEST, POSITION_SOUTHEAST, ALIGN_CENTER, \
    CURSOR_ARROW
from pygame_menu._scrollarea import get_scrollbars_from_position
from pygame_menu.utils import assert_alignment, assert_cursor, assert_vector, \
    assert_position, assert_color, format_color, assert_position_vector
from pygame_menu.widgets import HighlightSelection, NoneSelection, MENUBAR_STYLE_ADAPTIVE, \
    MENUBAR_STYLE_SIMPLE, MENUBAR_STYLE_TITLE_ONLY, MENUBAR_STYLE_TITLE_ONLY_DIAGONAL, \
    MENUBAR_STYLE_NONE, MENUBAR_STYLE_UNDERLINE, MENUBAR_STYLE_UNDERLINE_TITLE
from pygame_menu.widgets.core import Selection
from pygame_menu.widgets.core.widget import WidgetBorderPositionType, WIDGET_FULL_BORDER, \
    WIDGET_SHADOW_TYPE_ELLIPSE, WIDGET_SHADOW_TYPE_RECTANGULAR

from pygame_menu._types import ColorType, ColorInputType, Tuple, List, Union, Dict, \
    Any, Tuple2IntType, VectorInstance, Tuple2NumberType, NumberType, PaddingType, \
    Optional, Type, NumberInstance, PaddingInstance, Tuple3IntType, CursorType

TRANSPARENT_COLOR: ColorType = (0, 0, 0, 0)


def _check_menubar_style(style: int) -> bool:
    """
    Check menubar style.

    :param style: Style
    :return: ``True`` if correct
    """
    return style in (MENUBAR_STYLE_ADAPTIVE, MENUBAR_STYLE_SIMPLE, MENUBAR_STYLE_TITLE_ONLY,
                     MENUBAR_STYLE_TITLE_ONLY_DIAGONAL, MENUBAR_STYLE_NONE,
                     MENUBAR_STYLE_UNDERLINE, MENUBAR_STYLE_UNDERLINE_TITLE)


class Theme(object):
    """
    Class defining the visual rendering of menus and widgets.

    .. note::

        All colors must be defined with a tuple of 3 or 4 numbers in the formats:

            - (R, G, B)
            - (R, G, B, A)

        Red (R), Green (G), and Blue (B) must be numbers between ``0`` and ``255``.
        A means the alpha channel (opacity), if ``0`` the color is transparent, ``100`` means opaque.

    .. note::

        Themes only modify visual behaviour of the Menu. For other options
        like rows/columns, enabling or disabling overflow, position, or Menu
        width/height see Menu parameters.

    :param background_color: Menu background color
    :type background_color: tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`
    :param border_color: Menu border color. If border is an image, it will be split in 9 tiles to use top, left, bottom, right, and the corners
    :type border_color: tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`, None
    :param border_width: Border width in px. Used only if ``border_color`` is not an image
    :type border_width: int
    :param cursor_color: Cursor color (used in some text-gathering widgets like ``TextInput``)
    :type cursor_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param cursor_selection_color: Color of the text selection if the cursor is enabled on certain widgets
    :type cursor_selection_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param cursor_switch_ms: Interval of cursor switch between off and on status
    :type cursor_switch_ms: int, float
    :param focus_background_color: Color of the widget focus, this must be a tuple of 4 elements (R, G, B, A)
    :type focus_background_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param fps: Menu max fps (frames per second). If ``0`` there's no limit
    :type fps: int, float
    :param readonly_color: Color of the widget in readonly mode
    :type readonly_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param readonly_selected_color: Color of the selected widget in readonly mode
    :type readonly_selected_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param scrollarea_outer_margin: Outer ScrollArea margin in px; the tuple is added to computed ScrollArea width/height, it can add a margin to bottom/right scrolls after widgets. If value less than ``1`` use percentage of width/height. It cannot be a negative value
    :type scrollarea_outer_margin: tuple, list
    :param scrollarea_position: Position of ScrollArea scrollbars. See :py:mod:`pygame_menu.locals`
    :type scrollarea_position: str
    :param scrollbar_color: Scrollbars color
    :type scrollbar_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param scrollbar_cursor: Scrollbar cursor if mouse is placed over. If ``None`` the scrollbar don't change the cursor
    :type scrollbar_cursor: int, :py:class:`pygame.cursors.Cursor`, None
    :param scrollbar_shadow: Indicate if a shadow is drawn on each scrollbar
    :type scrollbar_shadow: bool
    :param scrollbar_shadow_color: Color of the scrollbar shadow
    :type scrollbar_shadow_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param scrollbar_shadow_offset: Offset of the scrollbar shadow
    :type scrollbar_shadow_offset: int
    :param scrollbar_shadow_position: Position of the scrollbar shadow. See :py:mod:`pygame_menu.locals`
    :type scrollbar_shadow_position: str
    :param scrollbar_slider_color: Color of the sliders
    :type scrollbar_slider_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param scrollbar_slider_hover_color: Color of the slider if hovered or clicked
    :type scrollbar_slider_hover_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param scrollbar_slider_pad: Space between slider and scrollbars borders
    :type scrollbar_slider_pad: int, float
    :param scrollbar_thick: Scrollbar thickness in px
    :type scrollbar_thick: int
    :param selection_color: Color of the selected widget. It updates both selected font and ``widget_selection_effect`` color
    :type selection_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param surface_clear_color: Surface clear color before applying background function
    :type surface_clear_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param title: Title is enabled/disabled. If disabled the object is ``hidden``
    :type title: bool
    :param title_background_color: Title background color
    :type title_background_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param title_bar_modify_scrollarea: If ``True`` title bar modifies the scrollbars of the scrollarea depending on the style
    :type title_bar_modify_scrollarea: bool
    :param title_bar_style: Style of the title, use :py:class:`pygame_menu.widgets.MenuBar` widget styles
    :type title_bar_style: int
    :param title_close_button: Draw a back-box button on header to close the Menu. If user moves through nested submenus this buttons turns to a back-arrow
    :type title_close_button: bool
    :param title_close_button_background_color: Title back-box background color
    :type title_close_button_background_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param title_close_button_cursor: Cursor applied over title close button
    :type title_close_button_cursor: int, :py:class:`pygame.cursors.Cursor`, None
    :param title_fixed: If ``True`` title is drawn over the scrollarea, forcing widget surface area to be drawn behind the title
    :type title_fixed: bool
    :param title_floating: If ``True`` title don't contribute height to the Menu. Thus, scroll uses full menu width/height
    :type title_floating: bool
    :param title_font: Title font
    :type title_font: str, :py:class:`pygame.font.Font`, :py:class:`pathlib.Path`
    :param title_font_antialias: Title font renders with antialiasing
    :type title_font_antialias: bool
    :param title_font_color: Title font color
    :type title_font_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param title_font_shadow: Enable title font shadow
    :type title_font_shadow: bool
    :param title_font_shadow_color: Title font shadow color
    :type title_font_shadow_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param title_font_shadow_offset: Offset of title font shadow in px
    :type title_font_shadow_offset: int
    :param title_font_shadow_position: Position of the title font shadow. See :py:mod:`pygame_menu.locals`
    :type title_font_shadow_position: str
    :param title_font_size: Font size of the title
    :type title_font_size: int
    :param title_offset: Offset (x-position, y-position) of title in px
    :type title_offset: tuple, list
    :param title_updates_pygame_display: If ``True`` the menu title updates See :py:mod:`pygame.display.caption` automatically on draw
    :type title_updates_pygame_display: bool
    :param widget_alignment: Widget default `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_. See :py:mod:`pygame_menu.locals`
    :type widget_alignment: str
    :param widget_alignment_ignore_scrollbar_thickness: Widget positioning ignores the scrollbar thickness. If ``True``, the widgets only consider the menu size, ignoring the thickness of the visible scrollbars
    :type widget_alignment_ignore_scrollbar_thickness: bool
    :param widget_background_color: Background color of a widget, it can be a color, ``None`` (transparent), or a BaseImage object. Background fills the entire widget + the padding
    :type widget_background_color: tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`, None
    :param widget_background_inflate: Inflate background on x-axis and y-axis (x, y) in px. By default, it uses the highlight margin. This parameter is visual only. For modifying widget size use padding instead
    :type widget_background_inflate: tuple, list
    :param widget_background_inflate_to_selection: If ``True`` widget will inflate to match selection effect margin and overrides ``widget_background_inflate``
    :type widget_background_inflate_to_selection: bool
    :param widget_border_color: Widget border color
    :type widget_border_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param widget_border_inflate: Widget inflate size on x-axis and y-axis (x, y) in px. These values cannot be negative
    :type widget_border_inflate: tuple, list
    :param widget_border_position: Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
    :type widget_border_position: str, tuple, list
    :param widget_border_width: Widget border width in px. If ``0`` the border is disabled. Border width don't contribute to the widget width/height, it's visual-only
    :type widget_border_width: int
    :param widget_box_arrow_color: Widget box arrow color, used by some widgets such as DropSelect, Fancy Selector, etc.
    :type widget_box_arrow_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param widget_box_arrow_margin: Widget box arrow margin (left, right, vertical) in px, used by some widgets such as DropSelect, Fancy Selector, etc.
    :type widget_box_arrow_margin: tuple
    :param widget_box_background_color: Widget box background color, used by some widgets such as DropSelect, Fancy Selector, etc.
    :type widget_box_background_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param widget_box_border_color: Widget box border color, used by some widgets such as DropSelect, Fancy Selector, etc.
    :type widget_box_border_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param widget_box_border_width: Widget box border width in px, used by some widgets such as DropSelect, Fancy Selector, etc.
    :type widget_box_border_width: int
    :param widget_box_inflate: Widget box inflate on x-axis and y-axis (x, y) in px, used by some widgets such as DropSelect, Fancy Selector, etc.
    :type widget_box_inflate: tuple, list
    :param widget_box_margin: Box margin on x-axis and y-axis (x, y) in px
    :type widget_box_margin: tuple, list
    :param widget_cursor: Widget cursor if mouse is placed over. If ``None`` the widget don't change the cursor
    :type widget_cursor: int, :py:class:`pygame.cursors.Cursor`, None
    :param widget_font: Widget font path or name
    :type widget_font: str, :py:class:`pygame.font.Font`, :py:class:`pathlib.Path`
    :param widget_font_antialias: Widget font renders with antialiasing
    :type widget_font_antialias: bool
    :param widget_font_background_color: Widget font background color. If ``None`` the value will be the same as ``background_color`` if it is a color object and if ``widget_font_background_color_from_menu`` is ``True`` and ``widget_background_color`` is ``None``
    :type widget_font_background_color: tuple, list, str, int, :py:class:`pygame.Color`, None
    :param widget_font_background_color_from_menu: Use Menu background color as font background color. Disabled by default
    :type widget_font_background_color_from_menu: bool
    :param widget_font_color: Color of the font
    :type widget_font_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param widget_font_shadow: Indicate if the widget font shadow is enabled
    :type widget_font_shadow: bool
    :param widget_font_shadow_color: Color of the widget font shadow
    :type widget_font_shadow_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param widget_font_shadow_offset: Offset of the widget font shadow in px
    :type widget_font_shadow_offset: int
    :param widget_font_shadow_position: Position of the widget font shadow. See :py:mod:`pygame_menu.locals`
    :type widget_font_shadow_position: str
    :param widget_font_size: Font size
    :type widget_font_size: int
    :param widget_margin: Horizontal and vertical margin of each element in Menu in px
    :type widget_margin: tuple, list
    :param widget_padding: Padding of the widget according to CSS rules. It can be a single digit, or a tuple of 2, 3, or 4 elements. Padding modifies widget width/height
    :type widget_padding: int, float, tuple, list
    :param widget_offset: x-axis and y-axis (x, y) offset of widgets within Menu in px respect to top-left corner. If value less than ``1`` use percentage of width/height. It cannot be a negative value
    :type widget_offset: tuple, list
    :param widget_selection_effect: Widget selection effect object. This is visual-only, the selection properties does not affect widget height/width
    :type widget_selection_effect: :py:class:`pygame_menu.widgets.core.Selection`, None
    :param widget_shadow_aa: Widget shadow antialiasing factor, default is x4
    :type widget_shadow_aa: int
    :param widget_shadow_color: The color of the widget shadow
    :type widget_shadow_color: tuple, list, str, int, :py:class:`pygame.Color`
    :param widget_shadow_radius: The border radius of the widget shadow
    :type widget_shadow_radius: int
    :param widget_shadow_type: The shadow type, it can be 'rectangular' or 'ellipse'
    :type widget_shadow_type: str
    :param widget_shadow_width: The width of the shadow in px
    :type widget_shadow_width: int
    :param widget_tab_size: Widget tab size
    :type widget_tab_size: int
    :param widget_url_color: Color of url text links
    :type widget_url_color: tuple, list, str, int, :py:class:`pygame.Color`
    """
    _disable_validation: bool
    background_color: Union[ColorType, 'BaseImage']
    border_color: Union[ColorType, 'BaseImage']
    cursor_color: ColorType
    cursor_selection_color: ColorType
    cursor_switch_ms: NumberType
    focus_background_color: ColorType
    fps: NumberType
    readonly_color: ColorType
    readonly_selected_color: ColorType
    scrollarea_outer_margin: Tuple2NumberType
    scrollarea_position: str
    scrollbar_color: ColorType
    scrollbar_cursor: CursorType
    scrollbar_shadow: bool
    scrollbar_shadow_color: ColorType
    scrollbar_shadow_offset: int
    scrollbar_shadow_position: str
    scrollbar_slider_color: ColorType
    scrollbar_slider_hover_color: ColorType
    scrollbar_slider_pad: NumberType
    scrollbar_thick: int
    selection_color: ColorType
    surface_clear_color: ColorType
    title: bool
    title_background_color: ColorType
    title_bar_modify_scrollarea: bool
    title_bar_style: int
    title_close_button: bool
    title_close_button_background_color: ColorType
    title_close_button_cursor: CursorType
    title_fixed: bool
    title_floating: bool
    title_font: FontType
    title_font_antialias: bool
    title_font_color: ColorType
    title_font_shadow: bool
    title_font_shadow_color: ColorType
    title_font_shadow_offset: int
    title_font_shadow_position: str
    title_font_size: int
    title_offset: Tuple2NumberType
    title_updates_pygame_display: bool
    widget_alignment: str
    widget_alignment_ignore_scrollbar_thickness: bool
    widget_background_color: Optional[Union[ColorType, 'BaseImage']]
    widget_background_inflate: Tuple2IntType
    widget_background_inflate_to_selection: bool
    widget_border_color: ColorType
    widget_border_inflate: Tuple2IntType
    widget_border_position: WidgetBorderPositionType
    widget_border_width: int
    widget_box_arrow_color: ColorType
    widget_box_arrow_margin: Tuple3IntType
    widget_box_background_color: ColorType
    widget_box_border_color: ColorType
    widget_box_border_width: int
    widget_box_inflate: Tuple2IntType
    widget_box_margin: Tuple2NumberType
    widget_cursor: CursorType
    widget_font: FontType
    widget_font_antialias: str
    widget_font_background_color: Optional[ColorType]
    widget_font_background_color_from_menu: bool
    widget_font_color: ColorType
    widget_font_shadow: bool
    widget_font_shadow_color: ColorType
    widget_font_shadow_offset: NumberType
    widget_font_shadow_position: str
    widget_font_size: int
    widget_margin: Tuple2NumberType
    widget_offset: Tuple2NumberType
    widget_padding: PaddingType
    widget_selection_effect: 'pygame_menu.widgets.core.Selection'
    widget_shadow_aa: int
    widget_shadow_color: ColorType
    widget_shadow_radius: int
    widget_shadow_type: str
    widget_shadow_width: int
    widget_tab_size: int
    widget_url_color: ColorType

    def __init__(self, **kwargs) -> None:

        # Menu general
        self.background_color = self._get(kwargs, 'background_color', 'color_image', (220, 220, 220))
        self.border_color = self._get(kwargs, 'border_color', 'color_image_none')
        self.border_width = self._get(kwargs, 'border_width', int, 0)
        self.focus_background_color = self._get(kwargs, 'focus_background_color', 'color', (0, 0, 0, 180))
        self.fps = self._get(kwargs, 'fps', NumberInstance, 30)
        self.readonly_color = self._get(kwargs, 'readonly_color', 'color', (120, 120, 120))
        self.readonly_selected_color = self._get(kwargs, 'readonly_selected_color', 'color', (190, 190, 190))
        self.selection_color = self._get(kwargs, 'selection_color', 'color', (255, 255, 255))
        self.surface_clear_color = self._get(kwargs, 'surface_clear_color', 'color', (0, 0, 0))

        # Cursor/Text gathering
        self.cursor_color = self._get(kwargs, 'cursor_color', 'color', (0, 0, 0))
        self.cursor_selection_color = self._get(kwargs, 'cursor_selection_color', 'color', (30, 30, 30, 120))
        self.cursor_switch_ms = self._get(kwargs, 'cursor_switch_ms', NumberInstance, 750)

        # Menubar/Title
        self.title = self._get(kwargs, 'title', bool, True)
        self.title_background_color = self._get(kwargs, 'title_background_color', 'color', (70, 70, 70))
        self.title_bar_modify_scrollarea = self._get(kwargs, 'title_bar_modify_scrollarea', bool, True)
        self.title_bar_style = self._get(kwargs, 'title_bar_style', int, MENUBAR_STYLE_ADAPTIVE)
        self.title_close_button = self._get(kwargs, 'title_close_button', bool, True)
        self.title_close_button_background_color = self._get(kwargs, 'title_close_button_background_color', 'color', (255, 255, 255))
        self.title_close_button_cursor = self._get(kwargs, 'title_close_button_cursor', 'cursor')
        self.title_fixed = self._get(kwargs, 'title_fixed', bool, True)
        self.title_floating = self._get(kwargs, 'title_floating', bool, False)
        self.title_font = self._get(kwargs, 'title_font', 'font', FONT_OPEN_SANS)
        self.title_font_antialias = self._get(kwargs, 'title_font_antialias', bool, True)
        self.title_font_color = self._get(kwargs, 'title_font_color', 'color', (220, 220, 220))
        self.title_font_shadow = self._get(kwargs, 'title_font_shadow', bool, False)
        self.title_font_shadow_color = self._get(kwargs, 'title_font_shadow_color', 'color', (0, 0, 0))
        self.title_font_shadow_offset = self._get(kwargs, 'title_font_shadow_offset', int, 2)
        self.title_font_shadow_position = self._get(kwargs, 'title_font_shadow_position', 'position', POSITION_NORTHWEST)
        self.title_font_size = self._get(kwargs, 'title_font_size', int, 40)
        self.title_offset = self._get(kwargs, 'title_offset', 'tuple2', (5, -1))
        self.title_updates_pygame_display = self._get(kwargs, 'title_updates_pygame_display', bool, False)

        # ScrollArea
        self.scrollarea_outer_margin = self._get(kwargs, 'scrollarea_outer_margin', 'tuple2', (0, 0))
        self.scrollarea_position = self._get(kwargs, 'scrollarea_position', str, POSITION_SOUTHEAST)

        # ScrollBar
        self.scrollbar_color = self._get(kwargs, 'scrollbar_color', 'color', (235, 235, 235))
        self.scrollbar_cursor = self._get(kwargs, 'scrollbar_cursor', 'cursor')
        self.scrollbar_shadow = self._get(kwargs, 'scrollbar_shadow', bool, False)
        self.scrollbar_shadow_color = self._get(kwargs, 'scrollbar_shadow_color', 'color', (0, 0, 0))
        self.scrollbar_shadow_offset = self._get(kwargs, 'scrollbar_shadow_offset', int, 2)
        self.scrollbar_shadow_position = self._get(kwargs, 'scrollbar_shadow_position', 'position', POSITION_NORTHWEST)
        self.scrollbar_slider_color = self._get(kwargs, 'scrollbar_slider_color', 'color', (200, 200, 200))
        self.scrollbar_slider_hover_color = self._get(kwargs, 'scrollbar_slider_hover_color', 'color', (170, 170, 170))
        self.scrollbar_slider_pad = self._get(kwargs, 'scrollbar_slider_pad', NumberInstance, 0)
        self.scrollbar_thick = self._get(kwargs, 'scrollbar_thick', int, 20)

        # Generic widget themes
        self.widget_alignment = self._get(kwargs, 'widget_alignment', 'alignment', ALIGN_CENTER)
        self.widget_alignment_ignore_scrollbar_thickness = self._get(kwargs, 'widget_alignment_ignore_scrollbar_thickness', bool, False)
        self.widget_background_color = self._get(kwargs, 'widget_background_color', 'color_image_none')
        self.widget_background_inflate = self._get(kwargs, 'background_inflate', 'tuple2int', (0, 0))
        self.widget_background_inflate_to_selection = self._get(kwargs, 'widget_background_inflate_to_selection', bool,
                                                                False)
        self.widget_border_color = self._get(kwargs, 'widget_border_color', 'color', (0, 0, 0))
        self.widget_border_inflate = self._get(kwargs, 'widget_border_inflate', 'tuple2int', (0, 0))
        self.widget_border_position = self._get(kwargs, 'widget_border_position', 'position_vector', WIDGET_FULL_BORDER)
        self.widget_border_width = self._get(kwargs, 'widget_border_width', int, 0)
        self.widget_box_arrow_color = self._get(kwargs, 'widget_box_arrow_color', 'color', (150, 150, 150))
        self.widget_box_arrow_margin = self._get(kwargs, 'widget_box_arrow_margin', 'tuple3int', (5, 5, 0))
        self.widget_box_background_color = self._get(kwargs, 'widget_box_background_color', 'color', (255, 255, 255))
        self.widget_box_border_color = self._get(kwargs, 'widget_box_border_color', 'color', (0, 0, 0))
        self.widget_box_border_width = self._get(kwargs, 'widget_box_border_width', int, 1)
        self.widget_box_inflate = self._get(kwargs, 'widget_box_inflate', 'tuple2int', (0, 0))
        self.widget_box_margin = self._get(kwargs, 'widget_box_margin', 'tuple2', (25, 0))
        self.widget_cursor = self._get(kwargs, 'widget_cursor', 'cursor', CURSOR_ARROW)
        self.widget_font = self._get(kwargs, 'widget_font', 'font', FONT_OPEN_SANS)
        self.widget_font_antialias = self._get(kwargs, 'widget_font_antialias', bool, True)
        self.widget_font_background_color = self._get(kwargs, 'widget_font_background_color', 'color_none', )
        self.widget_font_background_color_from_menu = self._get(kwargs, 'widget_font_background_color_from_menu', bool, False)
        self.widget_font_color = self._get(kwargs, 'widget_font_color', 'color', (70, 70, 70))
        self.widget_font_shadow = self._get(kwargs, 'widget_font_shadow', bool, False)
        self.widget_font_shadow_color = self._get(kwargs, 'widget_font_shadow_color', 'color', (0, 0, 0))
        self.widget_font_shadow_offset = self._get(kwargs, 'widget_font_shadow_offset', int, 2)
        self.widget_font_shadow_position = self._get(kwargs, 'widget_font_shadow_position', 'position', POSITION_NORTHWEST)
        self.widget_font_size = self._get(kwargs, 'widget_font_size', int, 30)
        self.widget_margin = self._get(kwargs, 'widget_margin', 'tuple2', (0, 0))
        self.widget_offset = self._get(kwargs, 'widget_offset', 'tuple2', (0, 0))
        self.widget_padding = self._get(kwargs, 'widget_padding', PaddingInstance, (4, 8))
        self.widget_selection_effect = self._get(kwargs, 'widget_selection_effect', Selection, HighlightSelection(margin_x=0, margin_y=0))
        self.widget_shadow_aa = self._get(kwargs, 'widget_shadow_aa', int, 4)
        self.widget_shadow_color = self._get(kwargs, 'widget_shadow_color', 'color', (0, 0, 0))
        self.widget_shadow_radius = self._get(kwargs, 'widget_shadow_radius', int, 0)
        self.widget_shadow_type = self._get(kwargs, 'widget_shadow_type', str, WIDGET_SHADOW_TYPE_RECTANGULAR)
        self.widget_shadow_width = self._get(kwargs, 'widget_shadow_width', int, 0)
        self.widget_tab_size = self._get(kwargs, 'widget_tab_size', int, 4)
        self.widget_url_color = self._get(kwargs, 'widget_url_color', 'color', (6, 69, 173))

        # Upon this, no more kwargs should exist, raise exception if there's more
        for invalid_keyword in kwargs.keys():
            raise ValueError(f'parameter Theme.{invalid_keyword} does not exist')

        # Test purpose only, if True disables any validation
        self._disable_validation = False

    def validate(self) -> 'Theme':
        """
        Validate the values of the theme. If there's an invalid parameter throws an
        ``AssertionError``.

        This function also converts all lists to tuples. This is done because lists
        are mutable.

        :return: Self reference
        """
        if self._disable_validation:
            return self

        # Boolean asserts
        assert isinstance(self.scrollbar_shadow, bool)
        assert isinstance(self.title_bar_modify_scrollarea, bool)
        assert isinstance(self.title_close_button, bool)
        assert isinstance(self.title_font_antialias, bool)
        assert isinstance(self.title_font_shadow, bool)
        assert isinstance(self.widget_font_antialias, bool)
        assert isinstance(self.widget_font_background_color_from_menu, bool)
        assert isinstance(self.widget_font_shadow, bool)

        # Value type checks
        assert_alignment(self.widget_alignment)
        assert_cursor(self.scrollbar_cursor)
        assert_cursor(self.title_close_button_cursor)
        assert_cursor(self.widget_cursor)
        assert_font(self.title_font)
        assert_font(self.widget_font)
        assert_position(self.scrollbar_shadow_position)
        assert_position(self.title_font_shadow_position)
        assert_position(self.widget_font_shadow_position)
        assert_position_vector(self.widget_border_position)

        assert _check_menubar_style(self.title_bar_style)
        assert get_scrollbars_from_position(self.scrollarea_position) is not None

        # Check selection effect if None
        if self.widget_selection_effect is None:
            self.widget_selection_effect = NoneSelection()

        assert isinstance(self.border_width, int)
        assert isinstance(self.cursor_switch_ms, NumberInstance)
        assert isinstance(self.fps, NumberInstance)
        assert isinstance(self.scrollbar_shadow_offset, int)
        assert isinstance(self.scrollbar_slider_pad, NumberInstance)
        assert isinstance(self.scrollbar_thick, int)
        assert isinstance(self.title, bool)
        assert isinstance(self.title_fixed, bool)
        assert isinstance(self.title_floating, bool)
        assert isinstance(self.title_font_shadow_offset, int)
        assert isinstance(self.title_font_size, int)
        assert isinstance(self.title_updates_pygame_display, bool)
        assert isinstance(self.widget_background_inflate_to_selection, bool)
        assert isinstance(self.widget_border_width, int)
        assert isinstance(self.widget_box_border_width, int)
        assert isinstance(self.widget_font_shadow_offset, int)
        assert isinstance(self.widget_font_size, int)
        assert isinstance(self.widget_padding, PaddingInstance)
        assert isinstance(self.widget_selection_effect, Selection)
        assert isinstance(self.widget_shadow_aa, int)
        assert isinstance(self.widget_shadow_radius, int)
        assert isinstance(self.widget_shadow_width, int)
        assert isinstance(self.widget_tab_size, int)

        # Format colors, this converts all color lists to tuples automatically,
        # if it is an image, return the same object
        self.background_color = self._format_color_opacity(self.background_color)
        self.border_color = self._format_color_opacity(self.border_color, none=True)
        self.cursor_color = self._format_color_opacity(self.cursor_color)
        self.cursor_selection_color = self._format_color_opacity(self.cursor_selection_color)
        self.focus_background_color = self._format_color_opacity(self.focus_background_color)
        self.readonly_color = self._format_color_opacity(self.readonly_color)
        self.readonly_selected_color = self._format_color_opacity(self.readonly_selected_color)
        self.scrollbar_color = self._format_color_opacity(self.scrollbar_color)
        self.scrollbar_shadow_color = self._format_color_opacity(self.scrollbar_shadow_color)
        self.scrollbar_slider_color = self._format_color_opacity(self.scrollbar_slider_color)
        self.scrollbar_slider_hover_color = self._format_color_opacity(self.scrollbar_slider_hover_color)
        self.selection_color = self._format_color_opacity(self.selection_color)
        self.surface_clear_color = self._format_color_opacity(self.surface_clear_color)
        self.title_background_color = self._format_color_opacity(self.title_background_color)
        self.title_close_button_background_color = self._format_color_opacity(self.title_close_button_background_color)
        self.title_font_color = self._format_color_opacity(self.title_font_color)
        self.title_font_shadow_color = self._format_color_opacity(self.title_font_shadow_color)
        self.widget_background_color = self._format_color_opacity(self.widget_background_color, none=True)
        self.widget_border_color = self._format_color_opacity(self.widget_border_color)
        self.widget_box_arrow_color = self._format_color_opacity(self.widget_box_arrow_color)
        self.widget_box_background_color = self._format_color_opacity(self.widget_box_background_color)
        self.widget_box_border_color = self._format_color_opacity(self.widget_box_border_color)
        self.widget_font_background_color = self._format_color_opacity(self.widget_font_background_color, none=True)
        self.widget_font_color = self._format_color_opacity(self.widget_font_color)
        self.widget_font_shadow_color = self._format_color_opacity(self.widget_font_shadow_color)
        self.widget_shadow_color = self._format_color_opacity(self.widget_shadow_color)
        self.widget_url_color = self._format_color_opacity(self.widget_url_color)

        # List to tuple
        self.scrollarea_outer_margin = self._vec_to_tuple(self.scrollarea_outer_margin, 2, NumberInstance)
        self.title_offset = self._vec_to_tuple(self.title_offset, 2, NumberInstance)
        self.widget_background_inflate = self._vec_to_tuple(self.widget_background_inflate, 2, int)
        self.widget_border_inflate = self._vec_to_tuple(self.widget_border_inflate, 2, int)
        self.widget_box_arrow_margin = self._vec_to_tuple(self.widget_box_arrow_margin, 3, int)
        self.widget_box_inflate = self._vec_to_tuple(self.widget_box_inflate, 2, int)
        self.widget_box_margin = self._vec_to_tuple(self.widget_box_margin, 2, NumberInstance)
        self.widget_margin = self._vec_to_tuple(self.widget_margin, 2, NumberInstance)
        if isinstance(self.widget_padding, VectorInstance):
            self.widget_padding = self._vec_to_tuple(self.widget_padding)
            assert 2 <= len(self.widget_padding) <= 4, \
                'widget padding tuple length must be 2, 3 or 4'
            for p in self.widget_padding:
                assert isinstance(p, NumberInstance), \
                    'each padding element must be numeric (integer or float)'
                assert p >= 0, \
                    'all padding elements must be equal or greater than zero'
        else:
            assert self.widget_padding >= 0, 'padding cannot be a negative number'
        self.widget_offset = self._vec_to_tuple(self.widget_offset, 2, NumberInstance)

        # Check sizes
        assert self.border_width >= 0, 'border width must be equal or greater than zero'
        assert self.scrollarea_outer_margin[0] >= 0 and self.scrollarea_outer_margin[1] >= 0, \
            'scroll area outer margin must be equal or greater than zero on both axis'
        assert self.widget_offset[0] >= 0 and self.widget_offset[1] >= 0, \
            'widget offset must be equal or greater than zero'
        assert self.widget_background_inflate[0] >= 0 and self.widget_background_inflate[1] >= 0, \
            'widget background inflate must be equal or greater than zero on both axis'
        assert self.widget_border_inflate[0] >= 0 and self.widget_border_inflate[1] >= 0, \
            'widget border inflate must be equal or greater than zero on both axis'
        assert self.widget_box_inflate[0] >= 0 and self.widget_box_inflate[1] >= 0, \
            'widget box inflate inflate must be equal or greater than zero on both axis'
        assert self.widget_shadow_width >= 0 and self.widget_shadow_radius >= 0, \
            'widget shadow width and radius must be equal or greater than zero'

        # Check shadow type
        assert self.widget_shadow_type in (WIDGET_SHADOW_TYPE_ELLIPSE, WIDGET_SHADOW_TYPE_RECTANGULAR), \
            'widget shadow type must be ellipse or rectangular'

        assert self.cursor_switch_ms > 0, 'cursor switch ms must be greater than zero'
        assert self.fps >= 0, 'fps must be equal or greater than zero'
        assert self.scrollbar_shadow_offset > 0, 'scrollbar shadow offset must be greater than zero'
        assert self.scrollbar_slider_pad >= 0, 'slider pad must be equal or greater than zero'
        assert self.scrollbar_thick > 0, 'scrollbar thickness must be greater than zero'
        assert self.title_font_size > 0, 'title font size must be greater than zero'
        assert self.widget_border_width >= 0, 'widget border width must be equal or greater than zero'
        assert self.widget_box_border_width >= 0, 'widget border box width must be equal or greater than zero'
        assert self.widget_font_shadow_offset > 0, 'widget font shadow offset must be greater than zero'
        assert self.widget_font_size > 0, 'widget font size must be greater than zero'
        assert self.widget_shadow_aa >= 1, 'widget shadow antialiasing must be equal or greater than one'
        assert self.widget_tab_size >= 0, 'widget tab size must be equal or greater than zero'

        # Color asserts
        assert self.focus_background_color[3] != 0, \
            'focus background color cannot be fully transparent, suggested opacity between 1 and 255'

        return self

    def set_background_color_opacity(self, opacity: float) -> 'Theme':
        """
        Modify the Menu background color with given opacity.

        :param opacity: Opacity value, from ``0`` (transparent) to ``1`` (opaque)
        :return: Self reference
        """
        self.background_color = assert_color(self.background_color)
        assert isinstance(opacity, NumberInstance)
        assert 0 <= opacity <= 1, 'opacity must be a number between 0 (transparent) and 1 (opaque)'
        self.background_color = (self.background_color[0], self.background_color[1], self.background_color[2], int(float(opacity) * 255))
        return self

    @staticmethod
    def _vec_to_tuple(obj: Union[Tuple, List], check_length: int = 0, check_instance: Type = Any) -> Tuple:
        """
        Return a tuple from a list or tuple object.

        :param obj: Object
        :param check_length: Check length if not zero
        :return: Tuple
        """
        if isinstance(obj, tuple):
            v = obj
        elif isinstance(obj, list):
            v = tuple(obj)
        else:
            if check_length == 0:
                raise ValueError('object is not a vector')
            raise ValueError(f'object is not a vector of length {check_length}')
        if check_length > 0:
            if len(v) != check_length:
                raise ValueError(f'object is not a vector of length {check_length}')
        if check_instance is not Any:
            for i in v:
                assert isinstance(i, check_instance), \
                    f'{i} element of tuple {v} is not {check_instance} instance'
        return v

    def copy(self) -> 'Theme':
        """
        Creates a deep copy of the object.

        :return: Copied theme
        """
        self.validate()
        return copy.deepcopy(self)

    def __copy__(self) -> 'Theme':
        """
        Copy method.

        :return: Copied theme
        """
        return self.copy()

    @staticmethod
    def _format_color_opacity(
        color: Optional[Union[ColorInputType, 'BaseImage']],
        none: bool = False
    ) -> Optional[Union[ColorType, 'BaseImage']]:
        """
        Adds opacity to a 3 channel color. (R,G,B) -> (R,G,B,A) if the color
        has not an alpha channel. Also updates the opacity to a number between
        ``0`` (transparent) and ``255`` (opaque).

        Color may be an Image, so if this is the case return the same object.

        - If the color is a list, return a tuple.
        - If the color is ``None``, return ``None`` if ``None`` is True.

        :param color: Color object
        :param none: If ``True`` Color can be ``None``
        :return: Color in the same format
        """
        if isinstance(color, BaseImage):
            return color
        if color is None and none:
            return color
        color = format_color(color)
        if isinstance(color, VectorInstance):
            assert_color(color)
            if len(color) == 4:
                if isinstance(color, tuple):
                    return color
                return tuple(color)
            elif len(color) == 3:
                color = color[0], color[1], color[2], 255
        else:
            raise ValueError(f'invalid color type {color}, only tuple or list are valid')
        return color

    # noinspection PyTypeChecker
    @staticmethod
    def _get(params: Dict[str, Any], key: str,
             allowed_types: Optional[Union[Type, str, List[Type], Tuple[Type, ...]]] = None,
             default: Any = None) -> Any:
        """
        Return a value from a dictionary.

        Custom types (str)
            -   alignment           – pygame-menu alignment (locals)
            -   callable            – Is callable type, same as ``"function"``
            -   color               – Check color
            -   color_image         – Color or :py:class:`pygame_menu.baseimage.BaseImage`
            -   color_image_none    – Color, :py:class:`pygame_menu.baseimage.BaseImage`, or None
            -   color_none          – Color or None
            -   cursor              – Cursor object (pygame)
            -   font                – Font type
            -   image               – Value must be ``BaseImage``
            -   none                – None only
            -   position            – pygame-menu position (locals)
            -   position_vector     – pygame-menu position (str or vector)
            -   tuple2              – Only valid numeric tuples ``(x, y)`` or ``[x, y]``
            -   tuple2int           – Only valid integer tuples ``(x, y)`` or ``[x, y]``
            -   tuple3              – Only valid numeric tuples ``(x, y, z)`` or ``[x, y, z]``
            -   tuple3int           – Only valid integer tuples ``(x, y, z)`` or ``[x, y, z]``
            -   type                – Type-class (bool, str, etc...)

        :param params: Parameters dictionary
        :param key: Key to look for
        :param allowed_types: List of allowed types
        :param default: Default value to return
        :return: The value associated to the key
        """
        value = params.pop(key, default)
        if allowed_types is not None:
            other_types = []  # Contain other types to check from
            if not isinstance(allowed_types, VectorInstance):
                allowed_types = (allowed_types,)
            for val_type in allowed_types:

                if val_type == 'alignment':
                    assert_alignment(value)

                elif val_type == callable or val_type == 'function' or val_type == 'callable':
                    assert callable(value), \
                        'value must be callable type'

                elif val_type == 'color':
                    value = assert_color(value)

                elif val_type == 'color_image':
                    if not isinstance(value, BaseImage):
                        value = assert_color(value)

                elif val_type == 'color_image_none':
                    if not (value is None or isinstance(value, BaseImage)):
                        value = assert_color(value)

                elif val_type == 'color_none':
                    if value is not None:
                        value = assert_color(value)

                elif val_type == 'cursor':
                    assert_cursor(value)

                elif val_type == 'font':
                    assert_font(value)

                elif val_type == 'image':
                    assert isinstance(value, BaseImage), \
                        'value must be BaseImage type'

                elif val_type == 'none':
                    assert value is None

                elif val_type == 'position':
                    assert_position(value)

                elif val_type == 'position_vector':
                    assert_position_vector(value)

                elif val_type == 'type':
                    assert isinstance(value, type), \
                        'value is not type-class'

                elif val_type == 'tuple2':
                    assert_vector(value, 2)

                elif val_type == 'tuple2int':
                    assert_vector(value, 2, int)

                elif val_type == 'tuple3':
                    assert_vector(value, 3)

                elif val_type == 'tuple3int':
                    assert_vector(value, 3, int)

                else:  # Unknown type
                    assert isinstance(val_type, type), \
                        f'allowed type "{val_type}" is not a type-class'
                    other_types.append(val_type)

            # Check other types
            if len(other_types) > 0:
                others = tuple(other_types)
                assert isinstance(value, others), \
                    f'Theme.{key} type shall be in {others} types (got {type(value)})'

        return value


THEME_DEFAULT = Theme()

THEME_DARK = Theme(
    background_color=(40, 41, 35),
    cursor_color=(255, 255, 255),
    cursor_selection_color=(80, 80, 80, 120),
    scrollbar_color=(39, 41, 42),
    scrollbar_slider_color=(65, 66, 67),
    scrollbar_slider_hover_color=(90, 89, 88),
    selection_color=(255, 255, 255),
    title_background_color=(47, 48, 51),
    title_font_color=(215, 215, 215),
    widget_font_color=(200, 200, 200)
)

THEME_BLUE = Theme(
    background_color=(228, 230, 246),
    scrollbar_shadow=True,
    scrollbar_slider_color=(150, 200, 230),
    scrollbar_slider_hover_color=(123, 173, 202),
    scrollbar_slider_pad=2,
    selection_color=(100, 62, 132),
    title_background_color=(62, 149, 195),
    title_font_color=(228, 230, 246),
    title_font_shadow=True,
    widget_font_color=(61, 170, 220)
)

THEME_GREEN = Theme(
    background_color=(186, 214, 177),
    scrollbar_slider_color=(125, 121, 114),
    scrollbar_slider_hover_color=(100, 96, 90),
    scrollbar_slider_pad=2,
    selection_color=(125, 121, 114),
    title_background_color=(125, 121, 114),
    title_font_color=(228, 230, 246),
    widget_font_color=(255, 255, 255)
)

THEME_ORANGE = Theme(
    background_color=(228, 100, 36),
    selection_color=(255, 255, 255),
    title_background_color=(170, 65, 50),
    widget_font_color=(0, 0, 0),
    widget_font_size=30
)

THEME_SOLARIZED = Theme(
    background_color=(239, 231, 211),
    cursor_color=(0, 0, 0),
    cursor_selection_color=(146, 160, 160, 120),
    selection_color=(207, 62, 132),
    title_background_color=(4, 47, 58),
    title_font_color=(38, 158, 151),
    widget_font_color=(102, 122, 130)
)



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - DYNAMIC BUTTON
Menu with dynamic buttons.
"""

__all__ = ['main']

import pygame_menu
from pygame_menu.examples import create_example_window

from random import randrange

surface = create_example_window('Example - Dynamic Button Append', (600, 400))
menu = pygame_menu.Menu(
    height=300,
    theme=pygame_menu.themes.THEME_BLUE,
    title='Welcome',
    width=400
)


def add_dynamic_button() -> 'pygame_menu.widgets.Button':
    """
    Append a button to the menu on demand.

    :return: Appended button
    """
    print(f'Adding a button dynamically, total: {len(menu.get_widgets()) - 2}')
    btn = menu.add.button(randrange(0, 10))

    def _update_button() -> None:
        count = btn.get_counter_attribute('count', 1, btn.get_title())
        btn.set_title(str(count))

    btn.update_callback(_update_button)
    return btn


menu.add.text_input('Name: ', default='John Doe')
menu.add.button('Play', add_dynamic_button)
menu.add.button('Quit', pygame_menu.events.EXIT)


def main(test: bool = False) -> None:
    """
    Main function.

    :param test: Indicate function is being tested
    """
    menu.mainloop(surface, disable_loop=test)


if __name__ == '__main__':
    main()
